# SPDX-License-Identifier: MIT
from ..fw.agx.initdata import *
from ..fw.agx.channels import ChannelInfo
from ..hw.uat import MemoryAttr

from construct import Container
from m1n1.constructutils import Ver

def build_iomappings(agx, chip_id):
    def iomap(phys, size, range_size, rw):
        off = phys & 0x3fff
        virt = agx.io_allocator.malloc(size + 0x4000 + off)
        agx.uat.iomap_at(0, virt, phys - off, size + off, AttrIndex=MemoryAttr.Device)
        return IOMapping(phys, virt + off, size, range_size, rw)

    # for t8103
    if chip_id == 0x8103:
        return [
            iomap(0x204d00000, 0x1c000, 0x1c000, 1), # Fender
            iomap(0x20e100000, 0x4000, 0x4000, 0), # AICTimer
            iomap(0x23b104000, 0x4000, 0x4000, 1), # AICSWInt
            iomap(0x204000000, 0x20000, 0x20000, 1), # RGX
            IOMapping(), # UVD
            IOMapping(), # unused
            IOMapping(), # DisplayUnderrunWA
            iomap(0x23b2e8000, 0x1000, 0x1000, 0), # AnalogTempSensorControllerRegs
            iomap(0x23bc00000, 0x1000, 0x1000, 1), # PMPDoorbell
            iomap(0x204d80000, 0x5000, 0x5000, 1), # MetrologySensorRegs
            iomap(0x204d61000, 0x1000, 0x1000, 1), # GMGIFAFRegs
            iomap(0x200000000, 0xd6400, 0xd6400, 1), # MCache registers
            IOMapping(), # AICBankedRegisters
            iomap(0x23b738000, 0x1000, 0x1000, 1), # PMGRScratch
            IOMapping(), # NIA Special agent idle register die 0
            IOMapping(), # NIA Special agent idle register die 1
            IOMapping(), # CRE registers
            IOMapping(), # Streaming codec registers
            IOMapping(), #
            IOMapping(), #
            IOMapping(),
            IOMapping(),
            IOMapping(),
            IOMapping(),
            IOMapping(),
        ]
    elif chip_id == 0x8112:
        return [
            iomap(0x204d00000, 0x14000, 0x14000, 1), # Fender
            iomap(0x20e100000, 0x4000, 0x4000, 0), # AICTimer
            iomap(0x23b0c4000, 0x4000, 0x4000, 1), # AICSWInt
            iomap(0x204000000, 0x20000, 0x20000, 1), # RGX
            IOMapping(), # UVD
            IOMapping(), # unused
            IOMapping(), # DisplayUnderrunWA
            iomap(0x23b2c0000, 0x1000, 0x1000, 0), # AnalogTempSensorControllerRegs
            IOMapping(), # PMPDoorbell
            iomap(0x204d80000, 0x8000, 0x8000, 1), # MetrologySensorRegs
            iomap(0x204d61000, 0x1000, 0x1000, 1), # GMGIFAFRegs
            iomap(0x200000000, 0xd6400, 0xd6400, 1), # MCache registers
            IOMapping(), # AICBankedRegisters
            IOMapping(), # PMGRScratch
            IOMapping(), # NIA Special agent idle register die 0
            IOMapping(), # NIA Special agent idle register die 1
            iomap(0x204e00000, 0x10000, 0x10000, 0), # CRE registers
            iomap(0x27d050000, 0x4000, 0x4000, 0), # Streaming codec registers
            iomap(0x23b3d0000, 0x1000, 0x1000, 0), #
            iomap(0x23b3c0000, 0x1000, 0x1000, 0), #
            IOMapping(),
            IOMapping(),
            IOMapping(),
            IOMapping(),
            IOMapping(),
        ]
    elif chip_id in (0x6000, 0x6001, 0x6002):
        mcc_cnt = {0x6002: 16, 0x6001: 8, 0x6000: 4}
        return [
            iomap(0x404d00000, 0x1c000, 0x1c000, 1), # Fender
            iomap(0x20e100000, 0x4000, 0x4000, 0), # AICTimer
            iomap(0x28e104000, 0x4000, 0x4000, 1), # AICSWInt
            iomap(0x404000000, 0x20000, 0x20000, 1), # RGX
            IOMapping(), # UVD
            IOMapping(), # unused
            IOMapping(), # DisplayUnderrunWA
            iomap(0x28e494000, 0x1000, 0x1000, 0), # AnalogTempSensorControllerRegs
            IOMapping(), # PMPDoorbell
            iomap(0x404d80000, 0x8000, 0x8000, 1), # MetrologySensorRegs
            iomap(0x204d61000, 0x1000, 0x1000, 1), # GMGIFAFRegs
            iomap(0x200000000, mcc_cnt[chip_id] * 0xd8000, 0xd8000, 1), # MCache registers
            IOMapping(), # AICBankedRegisters
            IOMapping(), # PMGRScratch
            iomap(0x2643c4000, 0x1000, 0x1000, 1), # NIA Special agent idle register die 0
            iomap(0x22643c4000, 0x1000, 0x1000, 1) if chip_id == 0x6002 else IOMapping(), # NIA Special agent idle register die 1
            IOMapping(), # CRE registers
            IOMapping(), # Streaming codec registers
            iomap(0x28e3d0000, 0x1000, 0x1000, 1),
            iomap(0x28e3c0000, 0x2000, 0x2000, 0),
        ]
    elif chip_id in (0x6020, 0x6021, 0x6022):
        mcc_cnt = {0x6022: 16, 0x6021: 8, 0x6020: 4}
        return [
            iomap(0x404d00000, 0x144000, 0x144000, 1), # Fender
            iomap(0x20e100000, 0x4000, 0x4000, 0), # AICTimer
            iomap(0x28e106000, 0x4000, 0x4000, 1), # AICSWInt
            iomap(0x404000000, 0x20000, 0x20000, 1), # RGX
            IOMapping(), # UVD
            IOMapping(), # unused
            IOMapping(), # DisplayUnderrunWA
            iomap(0x28e478000, 0x4000, 0x4000, 0), # AnalogTempSensorControllerRegs
            IOMapping(), # PMPDoorbell
            iomap(0x404e08000, 0x8000, 0x8000, 1), # MetrologySensorRegs
            IOMapping(), # GMGIFAFRegs
            iomap(0x200000000, mcc_cnt[chip_id] * 0xd8000, 0xd8000, 1), # MCache registers
            iomap(0x28e118000, 0x4000, 0x4000, 0), # AICBankedRegisters
            IOMapping(), # PMGRScratch
            IOMapping(), # NIA Special agent idle register die 0
            IOMapping(), # NIA Special agent idle register die 1
            IOMapping(), # CRE registers
            IOMapping(), # Streaming codec registers
            iomap(0x28e3d0000, 0x4000, 0x4000, 1),
            iomap(0x28e3c0000, 0x4000, 0x4000, 0),
            iomap(0x28e3d8000, 0x4000, 0x4000, 1),
            iomap(0x404eac000, 0x4000, 0x4000, 1),
            IOMapping(),
            IOMapping(),
            IOMapping(),
        ]

CHIP_INFO = {
    0x8103: Container(
        chip_id = 0x8103,
        min_sram_volt = 850,
        max_power = 19551,
        max_freq_mhz = 1278,
        unk_87c = -220,
        unk_8cc = 9880,
        unk_924 = [[0] * 8] * 8,
        unk_e48 = [[0] * 8] * 8,
        unk_e24 = 112,
        gpu_fast_die0_sensor_mask64 = 0x12,
        gpu_fast_die1_sensor_mask64 = 0,
        gpu_fast_die0_sensor_mask64_alt = 0x12,
        gpu_fast_die1_sensor_mask64_alt = 0,
        gpu_fast_die0_sensor_present = 0x01,
        shared1_tab = [
            -1, 0x7282, 0x50ea, 0x370a, 0x25be, 0x1c1f, 0x16fb
        ] + ([-1] * 9),
        shared1_a4 = 0xffff,
        shared2_tab = [0x800, 0x1555, -1, -1, -1, -1, -1, -1, 0, 0],
        shared2_unk_508 = 0xc00007,
        unk_3cf4 = [1000.0, 0, 0, 0, 0, 0, 0, 0],
        unk_3d14 = [45.0, 0, 0, 0, 0, 0, 0, 0],
        unk_3d34_0 = [0, 0, 0, 0, 0, 0, 0, 0],
        unk_118ec = None,
        hwdb_4e0 = 0,
        hwdb_534 = 0,
        num_cores = 8,
        gpu_core = 11,
        gpu_rev = 4,
        hwdb_ab8 = 0x48,
        hwdb_abc = 0x8,
        hwdb_b30 = 0,
        rel_max_powers = [0, 19, 26, 38, 60, 87, 100],
        shared2_t1_coef = None,
        shared3_tab = [0] * 16,
        shared3_unk = 0,
        sram_base = 0,
        sram_size = 0,
        rc_unk_54 = 0xffff,
        unk_hws2_0 = 0,
        unk_hws2_4 = [0] * 8,
        unk_hws2_24 = 0,
        tiling_control = 0xa041,
    ),
    0x6001: Container(
        chip_id = 0x6001,
        min_sram_volt = 790,
        max_power = 81415,
        max_freq_mhz = 1296,
        unk_87c = 900,
        unk_8cc = 11000,
        unk_924 = [[i, *([0] * 7)] for i in [
            9.838, 9.819, 9.826, 9.799,
            0, 0, 0, 0,
        ]],
        unk_e48 = [[i, *([0] * 7)] for i in [
            13, 13, 13, 13, 0, 0, 0, 0,
        ]],
        unk_e24 = 125,
        gpu_fast_die0_sensor_mask64 = 0x80808080,
        gpu_fast_die1_sensor_mask64 = 0,
        gpu_fast_die0_sensor_mask64_alt = 0x90909090,
        gpu_fast_die1_sensor_mask64_alt = 0,
        gpu_fast_die0_sensor_present = 0x0f,
        shared1_tab = [0xffff] * 16,
        shared1_a4 = 0xffff,
        shared2_tab = [-1, -1, -1, -1, 0x2aa, 0xaaa, -1, -1, 0, 0],
        shared2_unk_508 = 0xcc00001,
        unk_3cf4 = [1314.0, 1330.0, 1314.0, 1288.0, 0, 0, 0, 0],
        unk_3d14 = [21.0, 21.0, 22.0, 21.0, 0, 0, 0, 0],
        unk_3d34_0 = [0, 0, 0, 0, 0, 0, 0, 0],
        unk_118ec = [
            0, 1, 2,
            1, 1, 90, 75, 1, 1,
            1, 2, 90, 75, 1, 1,
            1, 1, 90, 75, 1, 1
        ],
        hwdb_4e0 = 4,
        hwdb_534 = 1,
        num_cores = 32,
        gpu_core = 13,
        gpu_rev = 5,
        hwdb_ab8 = 0x2084,
        hwdb_abc = 0x80,
        hwdb_b30 = 0,
        rel_max_powers = [0, 15, 20, 27, 36, 52, 100],
        shared2_t1_coef = None,
        shared3_tab = [0] * 16,
        shared3_unk = 0,
        sram_base = 0,
        sram_size = 0,
        rc_unk_54 = 0xffff,
        unk_hws2_0 = 0,
        unk_hws2_4 = [0] * 8,
        unk_hws2_24 = 0,
        tiling_control = 0xa540,
    ),
    0x6002: Container(
        chip_id = 0x6002,
        min_sram_volt = 790,
        max_power = 166743,
        max_freq_mhz = 1296,
        unk_87c = 900,
        unk_8cc = 11000,
        unk_924 = [[i, *([0] * 7)] for i in [
            9.838, 9.819, 9.826, 9.799,
            9.799, 9.826, 9.819, 9.838,
        ]],
        unk_c30 = 0,
        unk_e48 = [[i, *([0] * 7)] for i in [
            13, 13, 13, 13, 13, 13, 13, 13,
        ]],
        unk_e24 = 125,
        gpu_fast_die0_sensor_mask64 = 0x8080808080808080,
        gpu_fast_die1_sensor_mask64 = 0,
        gpu_fast_die0_sensor_mask64_alt = 0x9090909090909090,
        gpu_fast_die1_sensor_mask64_alt = 0,
        gpu_fast_die0_sensor_present = 0xff,
        shared1_tab = [0xffff] * 16,
        shared1_a4 = 0xffff,
        shared2_tab = [-1, -1, -1, -1, 0x2aa, 0xaaa, -1, -1, 0, 0],
        shared2_unk_508 = 0xcc00001,
        unk_3cf4 = [1244.0, 1260.0, 1242.0, 1214.0,
                    1072.0, 1066.0, 1044.0, 1042.0],
        unk_3d14 = [18.0, 18.0, 18.0, 17.0, 15.0, 15.0, 15.0, 14.0],
        unk_3d34_0 = [0, 0, 0, 0, 0, 0, 0, 0],
        unk_8924 = 0,
        unk_118ec = [
            0, 1, 2,
            1, 1, 90, 75, 1, 1,
            1, 2, 90, 75, 1, 1,
            1, 1, 90, 75, 1, 1
        ],
        hwdb_4e0 = 4,
        hwdb_534 = 1,
        num_cores = 64,
        gpu_core = 13,
        gpu_rev = 5,
        hwdb_ab8 = 0x2084,
        hwdb_abc = 0x80,
        hwdb_b30 = 0,
        rel_max_powers = [0, 15, 19, 25, 34, 50, 100],
        shared2_t1_coef = None,
        shared3_tab = [0] * 16,
        shared3_unk = 0,
        sram_base = 0,
        sram_size = 0,
        rc_unk_54 = 0xffff,
        unk_hws2_0 = 0,
        unk_hws2_4 = [0] * 8,
        unk_hws2_24 = 0,
        tiling_control = 0xa540,
    ),
    0x8112: Container(
        chip_id = 0x8112,
        min_sram_volt = 780,
        max_power = 22800,
        max_freq_mhz = 1398,
        unk_87c = 900,
        unk_8cc = 11000,
        unk_924 = [[
            0.0, 0.0, 0.0, 0.0,
            5.3, 0.0, 5.3, 5.3,
        ]] + ([[0] * 8] * 7),
        unk_e48 = [[
            0.0, 0.0, 0.0, 0.0,
            5.3, 0.0, 5.3, 5.3,
        ]] + ([[0] * 8] * 7),
        unk_e24 = 125,
        gpu_fast_die0_sensor_mask64 = 0x6800,
        gpu_fast_die1_sensor_mask64 = 0,
        gpu_fast_die0_sensor_mask64_alt = 0x6800,
        gpu_fast_die1_sensor_mask64_alt = 0,
        gpu_fast_die0_sensor_present = 0x02,
        shared1_tab = [0xffff] * 16,
        shared1_a4 = 0,
        shared2_tab = [-1, -1, -1, -1, -1, -1, -1, -1, 0xaa5aa, 0],
        shared2_unk_508 = 0xc00000,
        unk_3cf4 = [1920.0, 0, 0, 0, 0, 0, 0, 0],
        unk_3d14 = [74.0, 0, 0, 0, 0, 0, 0, 0],
        unk_3d34_0 = [0, 0, 0, 0, 0, 0, 0, 0],
        unk_118ec = None,
        hwdb_4e0 = 4,
        hwdb_534 = 0,
        num_cores = 10,
        gpu_core = 15,
        gpu_rev = 3,
        hwdb_ab8 = 0x2048,
        hwdb_abc = 0x4000,
        hwdb_b30 = 1,
        rel_max_powers = [0, 18, 27, 37, 52, 66, 82, 96, 100],
        shared2_t1_coef = 7200,
        shared2_t2 = [0xf07, 0x4c0, 0x6c0, 0x8c0, 0xac0, 0xc40, 0xdc0, 0xec0, 0xf80],
        shared2_t3_coefs = [None, 20.0, 28.0, 36.0, 44.0, 50.0, 56.0, 60.0, 63.0],
        shared2_t3_scales = [9, 3209, 10400],
        unk_hws2_0 = 0,
        unk_hws2_4 = [0] * 8,
        unk_hws2_24 = 0,
        sram_base = 0,
        sram_size = 0,
        shared3_unk = 5,
        shared3_tab = [
            10700, 10700, 10700, 10700,
            10700, 6000, 1000, 1000,
            1000, 10700, 10700, 10700,
            10700, 10700, 10700, 10700,
        ],
        rc_unk_54 = 0xffff,
        tiling_control = 0xa041,
    ),
    0x6021: Container(
        chip_id = 0x6021,
        min_sram_volt = 790,
        max_power = 95892,
        max_freq_mhz = 1398,
        unk_87c = 500,
        unk_8cc = 11000,
        unk_924 = [
            [0.0, 8.2, 0.0, 6.9, 6.9] + [0] * 11,
            [0.0, 0.0, 0.0, 6.9, 6.9] + [0] * 11,
            [0.0, 8.2, 0.0, 6.9, 0.0] + [0] * 11,
            [0.0, 0.0, 0.0, 6.9, 0.0] + [0] * 11,
        ] + ([[0] * 16] * 4),
        unk_e48 = [
            [0.0, 9.0, 0.0, 8.0, 8.0] + [0] * 11,
            [0.0, 0.0, 0.0, 8.0, 8.0] + [0] * 11,
            [0.0, 9.0, 0.0, 8.0, 0.0] + [0] * 11,
            [0.0, 0.0, 0.0, 8.0, 0.0] + [0] * 11,
        ] + ([[0] * 16] * 4),
        unk_e24 = 125,
        gpu_fast_die0_sensor_mask64 = 0x40005000c000d00,
        gpu_fast_die1_sensor_mask64 = 0,
        gpu_fast_die0_sensor_mask64_alt = 0x140015001d001d00,
        gpu_fast_die1_sensor_mask64_alt = 0,
        gpu_fast_die0_sensor_present = None,
        shared1_tab = [0xffff] * 16,
        shared1_a4 = 0,
        shared2_tab = [0x800, 0x1555, -1, -1, -1, -1, -1, -1, 0xaaaaa, 0],
        shared2_unk_508 = 0xc00007,
        unk_3cf4 = [1564.0, 1416.0, 1428.0, 1614.0, 0, 0, 0, 0],
        unk_3d14 = [42.0, 39.0, 39.0, 44.0, 0, 0, 0, 0],
        unk_3d34_0 = [547.0, 0.0, 293.0, 0.0, 547.0, 0.0, 293.0, 0.0],
        unk_118ec = [
            0, 2, 2,
            1, 1, 90, 75, 1, 1,
            1, 2, 90, 75, 1, 1,
            1, 2, 90, 75, 1, 1,
            1, 1, 90, 75, 1, 1,
        ],
        hwdb_4e0 = 4,
        hwdb_534 = 0,
        num_cores = 40,
        gpu_core = 17,
        gpu_rev = 3,
        hwdb_ab8 = None,
        hwdb_abc = None,
        hwdb_b30 = 0,
        rel_max_powers = [0, 19, 26, 36, 48, 63, 79, 91, 100],
        shared2_t1_coef = 11000,
        shared2_t2 = [0xf07, 0x4c0, 0x680, 0x8c0, 0xa80, 0xc40, 0xd80, 0xec0, 0xf40],
        shared2_t3_coefs = [None, 20.0, 27.0, 36.0, 43.0, 50.0, 55.0, 60.0, 62.0],
        shared2_t3_scales = [9, 3209, 10400],
        unk_hws2_0 = 700,
        unk_hws2_4 = [1.0, 0.8, 0.2, 0.9, 0.1, 0.25, 0.7, 0.9],
        unk_hws2_24 = 6,
        sram_base = 0x404d60000,
        sram_size = 0x20000,
        shared3_unk = 8,
        shared3_tab = [
            125, 125, 125, 125, 125, 125, 125, 125,
            7500, 125, 125, 125, 125, 125, 125, 125
        ],
        rc_unk_54 = 4000,
        tiling_control = 0x180340,
    ),
    0x6020: Container(
        chip_id = 0x6020,
        min_sram_volt = 790,
        max_power = 95892,
        max_freq_mhz = 1398,
        unk_87c = 500,
        unk_8cc = 11000,
        unk_924 = [
            [0.0, 8.2, 0.0, 6.9, 6.9] + [0] * 11,
            [0.0, 0.0, 0.0, 6.9, 6.9] + [0] * 11,
            [0.0, 8.2, 0.0, 6.9, 0.0] + [0] * 11,
            [0.0, 0.0, 0.0, 6.9, 0.0] + [0] * 11,
        ] + ([[0] * 16] * 4),
        unk_e48 = [
            [0.0, 9.0, 0.0, 8.0, 8.0] + [0] * 11,
            [0.0, 0.0, 0.0, 8.0, 8.0] + [0] * 11,
            [0.0, 9.0, 0.0, 8.0, 0.0] + [0] * 11,
            [0.0, 0.0, 0.0, 8.0, 0.0] + [0] * 11,
        ] + ([[0] * 16] * 4),
        unk_e24 = 125,
        gpu_fast_die0_sensor_mask64 = 0xc000d00,
        gpu_fast_die1_sensor_mask64 = 0,
        gpu_fast_die0_sensor_mask64_alt = 0x1d001d00,
        gpu_fast_die1_sensor_mask64_alt = 0,
        gpu_fast_die0_sensor_present = None,
        shared1_tab = [0xffff] * 16,
        shared1_a4 = 0,
        shared2_tab = [0x800, 0x1555, -1, -1, -1, -1, -1, -1, 0xaaaaa, 0],
        shared2_unk_508 = 0xc00007,
        unk_3cf4 = [1564.0, 1416.0, 1428.0, 1614.0, 0, 0, 0, 0],
        unk_3d14 = [42.0, 39.0, 39.0, 44.0, 0, 0, 0, 0],
        unk_3d34_0 = [547.0, 0.0, 293.0, 0.0, 547.0, 0.0, 293.0, 0.0],
        unk_118ec = [
            0, 2, 2,
            1, 1, 90, 75, 1, 1,
            1, 2, 90, 75, 1, 1,
            1, 2, 90, 75, 1, 1,
            1, 1, 90, 75, 1, 1,
        ],
        hwdb_4e0 = 4,
        hwdb_534 = 0,
        num_cores = 20,
        gpu_core = 16,
        gpu_rev = 3,
        hwdb_ab8 = None,
        hwdb_abc = None,
        hwdb_b30 = 0,
        rel_max_powers = [0, 19, 26, 36, 48, 63, 79, 91, 100],
        shared2_t1_coef = 11000,
        shared2_t2 = [0xf07, 0x4c0, 0x680, 0x8c0, 0xa80, 0xc40, 0xd80, 0xec0, 0xf40],
        shared2_t3_coefs = [None, 20.0, 27.0, 36.0, 43.0, 50.0, 55.0, 60.0, 62.0],
        shared2_t3_scales = [9, 3209, 10400],
        unk_hws2_0 = 700,
        unk_hws2_4 = [1.0, 0.8, 0.2, 0.9, 0.1, 0.25, 0.7, 0.9],
        unk_hws2_24 = 6,
        sram_base = 0x404d60000,
        sram_size = 0x20000,
        shared3_unk = 8,
        shared3_tab = [
            125, 125, 125, 125, 125, 125, 125, 125,
            7500, 125, 125, 125, 125, 125, 125, 125
        ],
        rc_unk_54 = 4000,
        tiling_control = 0x180340,
    ),
}
def build_initdata(agx):
    sgx = agx.u.adt["/arm-io/sgx"]
    chosen = agx.u.adt["/chosen"]
    chip_info = CHIP_INFO[chosen.chip_id]
    agx.chip_info = chip_info

    initdata = agx.kshared.new(InitData)

    if Ver.check("V >= V13_3 && G == G14X"):
        initdata.ver_info = (0xb390, 0x70f8, 0x601, 0xb0)
    elif Ver.check("V >= V13_3 && G != G14X"):
        initdata.ver_info = (0x6ba0, 0x1f28, 0x601, 0xb0)

    initdata.regionA = agx.kshared.new_buf(0x4000, "InitData_RegionA").push()

    regionB = agx.kobj.new(InitData_RegionB)

    regionB.channels = agx.ch_info

    regionB.stats_ta = agx.kobj.new(InitData_GPUGlobalStatsTA).push()
    regionB.stats_3d = agx.kobj.new(InitData_GPUGlobalStats3D).push()

    # size: 0x180, Empty
    # 13.0: grew
    # 13.3: grew again
    #regionB.stats_cp = agx.kobj.new_buf(0x180, "RegionB.unkptr_180").push()
    regionB.stats_cp = agx.kobj.new_buf(0x980 + 0x800, "RegionB.stats_cp").push()

    # size: 0x3b80, few floats, few ints, needed for init
    regionB.hwdata_a = agx.kobj.new(AGXHWDataA(sgx, chip_info), track=False)

    # size: 0x80, empty
    regionB.fault_info = agx.kshared.new(AGXFaultInfo).push()

    # size: 0xc0, fw writes timestamps into this
    regionB.unk_198 = agx.kobj.new_buf(0xc0, "RegionB.unkptr_198").push()

    # size: 0xb80, io stuff
    hwdata = agx.kobj.new(AGXHWDataB(sgx, chip_info), track=False)
    hwdata.timestamp_region_base = agx.ktimestamp.start
    hwdata.io_mappings = build_iomappings(agx, chosen.chip_id)

    if chip_info.sram_base:
        virt = agx.io_allocator.malloc(chip_info.sram_size)
        hwdata.sgx_sram_ptr = virt
        agx.uat.iomap_at(0, virt, chip_info.sram_base, chip_info.sram_size,
                         AttrIndex=MemoryAttr.Shared)
    else:
        hwdata.sgx_sram_ptr = 0

    k = 1.02 #?
    count = sgx.perf_state_count
    table_count = sgx.perf_state_table_count
    base_pstate = sgx.getprop("gpu-perf-base-pstate", 3)
    base_freq = sgx.perf_states[base_pstate].freq
    max_freq = sgx.perf_states[count - 1].freq
    for i in range(count):
        ps = sgx.perf_states[i]
        hwdata.frequencies[i] = ps.freq // 1000000

        volt = [ps.volt] * 8
        for j in range(1, table_count):
            volt[j] = sgx.perf_states[count * j + i].volt
        sram_volt = [max(chip_info.min_sram_volt, i) for i in volt]

        hwdata.voltages[i] = volt
        hwdata.voltages_sram[i] = sram_volt

        regionB.hwdata_a.unk_74[i] = k
        hwdata.unk_9b4[i] = k
        hwdata.rel_max_powers[i] = chip_info.rel_max_powers[i]
        hwdata.rel_boost_freqs[i] = max(0, int((ps.freq - base_freq) / (max_freq - base_freq) * 100))

    cs_pstates = sgx.getprop("cs-perf-states", None)
    if cs_pstates:
        hwdata.cs_max_pstate = cs_pstates.count - 1
        hwdata.cs_frequencies = [(ps.freq // 1000000)
                                 for ps in cs_pstates.states[0]] + [0] * (16 - cs_pstates.count)
        hwdata.cs_voltages = [[(ps.volt // 1000), 0]
                              for ps in cs_pstates.states[0]] + [[0, 0]] * (16 - cs_pstates.count)
        hwdata.cs_voltages_sram = [[max(i[0], cs_pstates.min_sram_volt[0] // 1000) if i[0] else 0, 0] for i in hwdata.cs_voltages]
    else:
        hwdata.cs_max_pstate = 0
        hwdata.cs_frequencies = [0] * 16
        hwdata.cs_voltages = [[0, 0]] * 16
        hwdata.cs_voltages_sram = [[0, 0]] * 16

    afr_pstates = sgx.getprop("afr-perf-states", None)
    if afr_pstates:
        hwdata.afr_max_pstate = afr_pstates.count - 1
        hwdata.afr_frequencies = [(ps.freq // 1000000)
                                 for ps in afr_pstates.states[0]] + [0] * (8 - afr_pstates.count)
        hwdata.afr_voltages = [[(ps.volt // 1000), 0]
                              for ps in afr_pstates.states[0]] + [[0, 0]] * (8 - afr_pstates.count)
        hwdata.afr_voltages_sram = [[max(i[0], afr_pstates.min_sram_volt[0] // 1000) if i[0] else 0, 0] for i in hwdata.afr_voltages]
    else:
        hwdata.afr_max_pstate = 0
        hwdata.afr_frequencies = [0] * 8
        hwdata.afr_voltages = [[0, 0]] * 8
        hwdata.afr_voltages_sram = [[0, 0]] * 8

    regionB.hwdata_a.push()

    regionB.hwdata_b = hwdata.push()
    regionB.hwdata_b_addr2 = hwdata._addr

    regionB.fwlog_ring2 = agx.fwlog_ring

    # Unallocated, Size 0x1000
    regionB.unk_1b8 = agx.kobj.new_buf(0x1000, "RegionB.unkptr_1b8").push()

    # Unallocated, size 0x300
    regionB.unk_1c0 = agx.kobj.new_buf(0x300, "RegionB.unkptr_1c0").push()

    # Unallocated, unknown size
    regionB.unk_1c8 = agx.kobj.new_buf(0x1000, "RegionB.unkptr_1c8").push()

    # Size: 0x4000
    regionB.buffer_mgr_ctl = agx.kshared.new(InitData_BufferMgrCtl, track=True).push()

    agx.uat.iomap_at(0, 0x420000000, regionB.buffer_mgr_ctl._paddr_align, 0x4000,
                     AttrIndex=MemoryAttr.Shared, AP=0, UXN=1, PXN=1)

    regionB.buffer_mgr_ctl_gpu_addr = 0x420000000 + (regionB.buffer_mgr_ctl._addr & 0x3fff)

    regionB.unk_6a80 = 0
    regionB.gpu_idle = 0
    regionB.unk_6a9c = 0
    regionB.unk_ctr0 = 0
    regionB.unk_ctr1 = 0
    regionB.unk_6aa8 = 0
    regionB.unk_6aac = 0
    regionB.unk_ctr2 = 0
    regionB.unk_6ab4 = 0
    regionB.unk_6ab8 = 0
    regionB.unk_6abc = 0
    regionB.unk_6ac0 = 0
    regionB.unk_6ac4 = 0
    regionB.unk_ctr3 = 0
    regionB.unk_6acc = 0
    regionB.unk_6ad0 = 0
    regionB.unk_6ad4 = 0
    regionB.unk_6ad8 = 0
    regionB.unk_6adc = 0
    regionB.unk_6ae0 = 0
    regionB.unk_6ae4 = 0
    regionB.unk_6ae8 = 0
    regionB.unk_6aec = 0
    regionB.unk_6af0 = 0
    regionB.unk_ctr4 = 0
    regionB.unk_ctr5 = 0
    regionB.unk_6afc = 0

    initdata.regionB = regionB.push()

    initdata.regionC = agx.kshared.new(InitData_RegionC(sgx, chip_info), track=False).push()

    #self.regionC_addr = agx.ksharedshared_heap.malloc(0x88000)

    initdata.fw_status = agx.kobj.new(InitData_FWStatus)
    initdata.fw_status.fwctl_channel = agx.fwctl_chinfo
    initdata.fw_status.push()

    ## This section seems to be data that would be used by firmware side page allocation
    ## But the current firmware doesn't have this functionality enabled, so it's not used?
    initdata.uat_num_levels = 3
    initdata.uat_page_bits = 14
    initdata.uat_page_size = 0x4000

    if chip_info.chip_id in (0x8103, 0x8112):
        phys_mask = 0xffffffc000
    else:
        phys_mask = 0x3ffffffc000

    initdata.uat_level_info = [
        UatLevelInfo(36, 8, phys_mask),
        UatLevelInfo(25, 2048, phys_mask),
        UatLevelInfo(14, 2048, phys_mask),
    ]

    # Host handles FW allocations for existing firmware versions
    initdata.host_mapped_fw_allocations = 1


    #initdata.regionC.idle_ts = agx.u.mrs("CNTPCT_EL0") + 24000000
    #initdata.regionC.idle_unk = 0x5b2e8
    #initdata.regionC.idle_to_off_timeout_ms = 20000

    initdata.regionC.push()
    initdata.push()

    #print(InitData.parse_stream(agx.uat.iostream(0, initdata._addr)))
    return initdata



from m1n1.utils import Register8, Register16, Register32, RegMap, irange
from enum import IntEnum

class R_IRQ_MASK1(Register8):
    RING_PLUG   = 0
    RING_UNPLUG = 1
    TIP_PLUG    = 2
    TIP_UNPLUG  = 3

class R_IRQ_MASK3(Register8):
    HSDET_AUTO_DONE = 7

class E_DCID_GND_SEL(IntEnum):
    NONE = 0
    HS3  = 1
    HS4  = 2

class E_DCID_Z_RANGE(IntEnum):
    NONE = 0
    UNK2 = 2
    UNK3 = 3

class R_DCID_CTRL1(Register8):
    Z_RANGE = 2, 0, E_DCID_Z_RANGE

class R_DCID_CTRL2(Register8):
    GND_SEL = 6, 4, E_DCID_GND_SEL

class R_DCID_CTRL3(Register8):
    START = 0

class R_DCID_STATUS(Register32):
    OVERALL = 9, 0
    DONE    = 10
    U       = 20, 11
    D       = 30, 21

class E_DEBOUNCE_TIME(IntEnum):
    T_0MS   = 0b000
    T_125MS = 0b001
    T_250MS = 0b010
    T_500MS = 0b011
    T_750MS = 0b100
    T_1S    = 0b101

class R_TR_SENSE_CTRL(Register8):
    INV      = 7
    UNK1     = 6
    FALLTIME = 5, 3, E_DEBOUNCE_TIME
    RISETIME = 2, 0, E_DEBOUNCE_TIME

class R_TR_SENSE_STATUS(Register8):
    RING_PLUG   = 0
    RING_UNPLUG = 1
    TIP_PLUG    = 2
    TIP_UNPLUG  = 3

class R_HS_DET_STATUS2(Register8):
    HS_TRUE    = 1
    SHORT_TRUE = 0

class R_MSM_BLOCK_EN1(Register8):
    pass

class R_MSM_BLOCK_EN2(Register8):
    ASP_EN = 6
    BUS_EN = 5
    DAC_EN = 4
    ADC_EN = 3

class R_MSM_BLOCK_EN3(Register8):
    TR_SENSE_EN = 3
    DCID_EN     = 4

class R_HS_CLAMP_DISABLE(Register8):
    HS_CLAMP_DISABLE = 0

class E_SAMP_RATE(IntEnum):
    S_16KHZ   = 1
    S_24KHZ   = 2
    S_32KHZ   = 3
    S_48KHZ   = 4
    S_96KHZ   = 5
    S_192KHZ  = 6
    S_22K05HZ = 10
    S_44K1HZ  = 12
    S_88K2HZ  = 13
    S_176K4HZ = 14

class E_MCLK_SRC(IntEnum):
    RCO      = 0b00
    MCLK_PIN = 0b01
    BCLK     = 0b10
    PLL      = 0b11

class E_MCLK_FREQ(IntEnum):
    F_12MHZ     = 0b00
    F_24MHZ     = 0b01
    F_12_288KHZ = 0b10
    F_24_576KHZ = 0b11

class R_CCM_CTRL1(Register8):
    MCLK_SRC  = 1, 0, E_MCLK_SRC
    MCLK_FREQ = 3, 2, E_MCLK_FREQ

class E_REFCLK_DIV(IntEnum):
    DIV1 = 0b00
    DIV2 = 0b01
    DIV4 = 0b10
    DIV8 = 0b11

class R_CCM_CTRL3(Register8):
    REFCLK_DIV     = 2, 1, E_REFCLK_DIV
    REFCLK_IS_MCLK = 0 # BLCK otherwise

class R_CCM_CTRL4(Register8):
    REFCLK_EN = 0

class R_CCM_SAMP_RATE(Register8):
    RATE = 7, 0, E_SAMP_RATE

class E_PLL_MODE(IntEnum):
    UNSUPP      = 0b00
    BYPASS_512  = 0b01
    BYPASS_1024 = 0b10
    BYPASS_BOTH = 0b11

class R_PLL_CTRL(Register8):
    MODE = 2, 1, E_PLL_MODE
    EN   = 0

class E_WNF_CF(IntEnum):
    F_UNK   = 0b00
    F_300HZ = 0b11

class R_ADC_CTRL1(Register8):
    PREAMP_GAIN = 7, 6
    PGA_GAIN    = 5, 0

class R_ADC_CTRL4(Register8): # maybe
    WNF_CF = 5, 4, E_WNF_CF
    WNF_EN = 3

class R_DAC_CTRL1(Register8):
    UNMUTE  = 0

    HP_LOAD = 2 # maybe
    UNK1    = 3
    UNK2    = 4
    UNK3    = 5
    HIGH_V  = 6

class E_PULLDOWN_R(IntEnum):
    NONE      = 0x0
    R_UNK8    = 0x8
    R_1K1OHMS = 0xc

class R_DAC_CTRL2(Register8):
    PULLDOWN_R = 3, 0, E_PULLDOWN_R

class R_HP_VOL_CTRL(Register8):
    ZERO_CROSS  = 1
    SOFT        = 0

class E_BUS_SOURCE(IntEnum):
    EMPTY      = 0b0000
    ADC        = 0b0111
    ASP_RX_CH1 = 0b1101
    ASP_RX_CH2 = 0b1110

class R_BUS_DAC_SRC(Register8):
    CHB = 7, 4, E_BUS_SOURCE
    CHA = 3, 0, E_BUS_SOURCE

class R_BUS_ASP_TX_SRC(Register8):
    CH2 = 7, 4, E_BUS_SOURCE
    CH1 = 3, 0, E_BUS_SOURCE

class E_HSBIAS_SENSE_TRIP(IntEnum):
    C_12UA  = 0b000
    C_23UA  = 0b001
    C_41UA  = 0b010
    C_52UA  = 0b011
    C_64UA  = 0b100
    C_75UA  = 0b101
    C_93UA  = 0b110
    C_104UA = 0b111

class R_HSBIAS_SC_AUTOCTL(Register8):
    HSBIAS_SENSE_EN = 7
    AUTO_HSBIAS_HIZ = 6
    TIP_SENSE_EN    = 5
    SENSE_TRIP      = 2, 0, E_HSBIAS_SENSE_TRIP

class E_TIP_SENSE_CTRL(IntEnum):
    DISABLED  = 0b00
    DIG_INPUT = 0b01
    SHORT_DET = 0b11

class R_TIP_SENSE_CTRL2(Register8):
    CTRL = 7, 6, E_TIP_SENSE_CTRL
    INV  = 5

class E_HSBIAS_DET_MODE(IntEnum):
    DISABLED  = 0b00
    SHORT_DET = 0b01
    NORMAL    = 0b11

class E_HSBIAS_CTRL(IntEnum):
    HI_Z  = 0b00
    U_0V0 = 0b01
    U_2V0 = 0b10
    U_2V7 = 0b11

class R_MISC_DET_CTRL(Register8):
    UNK1            = 7
    DETECT_MODE     = 4, 3, E_HSBIAS_DET_MODE
    HSBIAS_CTRL     = 2, 1, E_HSBIAS_CTRL
    PDN_MIC_LVL_DET = 0

class E_S0_DEBOUNCE_TIME(IntEnum):
    T_10MS = 0b000
    T_20MS = 0b001
    T_30MS = 0b010
    T_40MS = 0b011
    T_50MS = 0b100
    T_60MS = 0b101
    T_70MS = 0b110
    T_80MS = 0b111

class R_MIC_DET_CTRL2(Register8):
    DEBOUNCE_TIME = 7, 5, E_S0_DEBOUNCE_TIME

class R_MIC_DET_CTRL4(Register8):
    LATCH_TO_VP = 1

class R_HS_DET_CTRL2(Register8):
    CTRL      = 7, 6
    SET       = 5, 4
    REF       = 3
    AUTO_TIME = 1, 0

class R_HS_SWITCH_CTRL(Register8):
    REF_HS3      = 7
    REF_HS4      = 6
    HSB_FILT_HS3 = 5
    HSB_FILT_HS4 = 4
    HSB_HS3      = 3
    HSB_HS4      = 2
    GNDHS_HS3    = 1
    GNDHS_HS4    = 0

class R_ASP_CTRL(Register8):
    TDM_MODE = 2
    BCLK_EN  = 1

class R_ASP_FSYNC_CTRL23(Register16):
    BCLK_PERIOD = 12, 1

class R_ASP_TX_HIZ_DLY_CTRL(Register8):
    DRV_Z     = 5, 4
    HIZ_DELAY = 3, 2
    FS        = 1
    UNK1      = 0

class R_ASP_RX_EN(Register8):
    CH2_EN = 1
    CH1_EN = 0

class R_ASP_CH_CTRL(Register32):
    WIDTH      = 23, 16
    SLOT_START = 10, 1
    EDGE       = 0 # set for rising edge

class CS42L84Regs(RegMap):
    DEVID  = irange(0x0, 5), Register8
    FREEZE = 0x6, Register8

    SW_RESET = 0x203, Register8

    IRQ_STATUS1     = 0x400, R_IRQ_MASK1
    IRQ_STATUS2     = 0x401, Register8
    IRQ_STATUS3     = 0x402, R_IRQ_MASK3
    PLL_LOCK_STATUS = 0x40e, Register8 # bit 0x10

    IRQ_MASK1  = 0x418, R_IRQ_MASK1
    IRQ_MASK2  = 0x419, Register8
    IRQ_MASK3  = 0x41a, R_IRQ_MASK3

    CCM_CTRL1        = 0x600, R_CCM_CTRL1
    CCM_SAMP_RATE    = 0x601, R_CCM_SAMP_RATE
    CCM_CTRL3        = 0x602, R_CCM_CTRL3
    CCM_CTRL4        = 0x603, R_CCM_CTRL4
    CCM_ASP_CLK_CTRL = 0x608, Register8

    PLL_CTRL     = 0x800, R_PLL_CTRL
    PLL_DIV_FRAC = irange(0x804, 3), Register8
    PLL_DIV_INT  = 0x807, Register8
    PLL_DIVOUT   = 0x808, Register8

    DCID_CTRL1          = 0x1200, R_DCID_CTRL1
    DCID_CTRL2          = 0x1201, R_DCID_CTRL2
    DCID_CTRL3          = 0x1202, R_DCID_CTRL3
    DCID_TRIM_OFFSET    = 0x1207, Register8
    DCID_TRIM_SLOPE     = 0x120a, Register8

    # R_pull = 1100 - (regval - 128)*2
    DCID_PULLDOWN_TRIM  = 0x120b, Register8
    DCID_STATUS         = 0x120c, R_DCID_STATUS

    # tip/ring sense
    TR_SENSE_CTRL1  = 0x1280, Register8
    TR_SENSE_CTRL2  = 0x1281, Register8
    RING_SENSE_CTRL = 0x1282, R_TR_SENSE_CTRL
    TIP_SENSE_CTRL  = 0x1283, R_TR_SENSE_CTRL
    TR_SENSE_STATUS = 0x1288, R_TR_SENSE_STATUS

    HSBIAS_SC_AUTOCTL = 0x1470, R_HSBIAS_SC_AUTOCTL
    WAKE_CTRL         = 0x1471, Register8
    TIP_SENSE_CTRL2   = 0x1473, R_TIP_SENSE_CTRL2
    MISC_DET_CTRL     = 0x1474, R_MISC_DET_CTRL
    MIC_DET_CTRL2     = 0x1478, R_MIC_DET_CTRL2
    MIC_DET_CTRL4     = 0x1477, R_MIC_DET_CTRL4

    HS_DET_STATUS1    = 0x147c, Register8
    HS_DET_STATUS2    = 0x147d, R_HS_DET_STATUS2
    HS_DET_IRQ_MASK   = irange(0x1480, 2), Register8
    HS_DET_IRQ_STATUS = irange(0x1484, 2), Register8

    MSM_BLOCK_EN1    = 0x1800, R_MSM_BLOCK_EN1
    MSM_BLOCK_EN2    = 0x1801, R_MSM_BLOCK_EN2
    MSM_BLOCK_EN3    = 0x1802, R_MSM_BLOCK_EN3

    HS_DET_CTRL1     = 0x1810, Register8
    HS_DET_CTRL2     = 0x1811, R_HS_DET_CTRL2
    HS_SWITCH_CTRL   = 0x1812, R_HS_SWITCH_CTRL
    HS_CLAMP_DISABLE = 0x1813, R_HS_CLAMP_DISABLE
    ADC_CTRL1        = 0x2000, R_ADC_CTRL1
    ADC_CTRL2        = 0x2001, Register8 # volume
    ADC_CTRL3        = 0x2002, Register8
    ADC_CTRL4        = 0x2003, R_ADC_CTRL4

    DAC_CTRL1     = 0x3000, R_DAC_CTRL1
    DAC_CTRL2     = 0x3001, R_DAC_CTRL2
    DACA_VOL_LSB  = 0x3004, Register8
    DACA_VOL_MSB  = 0x3005, Register8 # sign bit
    DACB_VOL_LSB  = 0x3006, Register8
    DACB_VOL_MSB  = 0x3007, Register8 # sign bit
    HP_VOL_CTRL   = 0x3020, R_HP_VOL_CTRL
    HP_CLAMP_CTRL = 0x3123, Register8

    BUS_ASP_TX_SRC = 0x4000, R_BUS_ASP_TX_SRC
    BUS_DAC_SRC    = 0x4001, R_BUS_DAC_SRC

    ASP_CTRL         = 0x5000, R_ASP_CTRL
    ASP_FSYNC_CTRL23 = 0x5010, R_ASP_FSYNC_CTRL23
    ASP_DATA_CTRL    = 0x5018, R_ASP_TX_HIZ_DLY_CTRL

    ASP_RX_EN = 0x5020, R_ASP_RX_EN
    ASP_TX_EN = 0x5024, Register8

    ASP_RX1_CTRL     = 0x5028, R_ASP_CH_CTRL # 32bit
    ASP_RX2_CTRL     = 0x502c, R_ASP_CH_CTRL # 32bit
    ASP_TX1_CTRL     = 0x5068, R_ASP_CH_CTRL
    ASP_TX2_CTRL     = 0x506c, R_ASP_CH_CTRL



# SPDX-License-Identifier: MIT
import datetime

from m1n1.constructutils import show_struct_trace, Ver
from m1n1.utils import *

Ver.set_version(hv.u)

from m1n1.trace.agx import AGXTracer
AGXTracer = AGXTracer._reloadcls(True)

agx_tracer = AGXTracer(hv, "/arm-io/gfx-asc", verbose=1)

agx_tracer.pause_after_init = True
agx_tracer.trace_usermap = False
agx_tracer.trace_kernmap = False
agx_tracer.trace_userva = False
agx_tracer.trace_kernva = False
agx_tracer.redump = True
agx_tracer.exclude_context_id = 1
agx_tracer.cmd_dump_dir = "gfxdumps"

agx_tracer.start()

def resume_tracing(ctx):
    fname = f"{datetime.datetime.now().isoformat()}.log"
    hv.set_logfile(open(f"gfxlogs/{fname}", "a"))
    agx_tracer.start()
    agx_tracer.resume()
    return True

def pause_tracing(ctx):
    agx_tracer.pause()
    agx_tracer.stop()
    hv.set_logfile(None)
    return True

hv.add_hvcall(100, resume_tracing)
hv.add_hvcall(101, pause_tracing)

trace_device("/arm-io/sgx", True)
#trace_device("/arm-io/mcc", True)
node = hv.adt["/arm-io/sgx"]
addr, size = node.get_reg(0)
hv.trace_range(irange(addr, 0x1000000), TraceMode.SYNC, name="sgx")
hv.trace_range(irange(addr + 0x2408000, 0x4000), TraceMode.OFF, name="sgx")



# SPDX-License-Identifier: MIT
from m1n1.constructutils import *
from construct import *
from .microsequence import *
from ...utils import RegMap, Register32

__all__ = []

class WorkCommandBarrier(ConstructClass):
    """
        sent before WorkCommand3D on the Submit3d queue.
        Might be for initializing the tile buckets?

    Example:
    00000004 0c378018 ffffffa0 00000c00 00000006 00000900 08002c9a 00000000
    00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    """
    subcon = Struct(
        "magic" / Const(0x4, Int32ul),
        "stamp_addr" / Int64ul,
        "stamp" / ROPointer(this.stamp_addr, StampCounter),
        "wait_value" / Int32ul,
        "event" / Int32ul, # Event number that signals a stamp check
        "stamp_self" / Int32ul,
        "uuid" / Int32ul,
        "unk" / Default(Int32ul, 0),
        Ver("G >= G14X", "pad" / ZPadding(0x20)),
    )

class WorkCommandInitBM(ConstructClass):
    """
        occasionally sent before WorkCommandTA on the SubmitTA queue.

    Example:
    00000004 0c378018 ffffffa0 00000c00 00000006 00000900 08002c9a 00000000
    00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    """
    subcon = Struct(
        "magic" / Const(0x6, Hex(Int32ul)),
        "context_id" / Hex(Int32ul), # Might be context?
        "buffer_mgr_slot" / Hex(Int32ul), # 0
        "unk_c" / Hex(Int32ul), # 0
        "unk_10" / Hex(Int32ul), # 0x30
        "buffer_mgr_addr" / Int64ul,
        "buffer_mgr" / ROPointer(this.buffer_mgr_addr, BufferManagerInfo),
        "stamp_value" / Hex(Int32ul),  # 0x100
    )

class WorkCommandComputeUnk10(ConstructClass):
    """
        occassionally sent before WorkCommandCP on the SubmitCP queue.
    """
    subcon = Struct(
        "magic" / Const(0xa, Hex(Int32ul)),
        "unk" / Hex(Int32ul),
    )

class WorkCommandComputeUnk11(ConstructClass):
    """
        occassionally sent before WorkCommandCP on the SubmitCP queue.
    """
    subcon = Struct(
        "magic" / Const(0xb, Hex(Int32ul)),
        "unk" / Hex(Int32ul),
    )

class Flag(ConstructValueClass):
    subcon = Hex(Int32ul)

    def __init__(self):
        self.value = 0

class LinkedListHead(ConstructClass):
    subcon = Struct(
        "prev" / Int64ul,
        "next" / Int64ul,
    )

    def __init__(self):
        super().__init__()
        self.prev = 0
        self.next = 0

class EventControlUnkBuf(ConstructValueClass):
    subcon = HexDump(Bytes(0x8))

    def __init__(self):
        super().__init__()
        self.value = b"\xff" * 8

class EventControl(ConstructClass):
    subcon = Struct(
        "event_count_addr" / Int64ul,
        "event_count" / ROPointer(this.event_count_addr, Int32ul),
        "submission_id" / Int32ul,
        "cur_count" / Int32ul,
        "unk_10" / Int32ul,
        "unk_14" / Int32ul,
        "unk_18" / Int64ul,
        "unk_20" / Int32ul,
        "vm_slot" / Int32ul,
        "has_ta" / Int32ul,
        "pstamp_ta" / Array(4, Int64ul),
        "has_3d" / Int32ul,
        "pstamp_3d" / Array(4, Int64ul),
        "has_cp" / Int32ul,
        "pstamp_cp" / Array(4, Int64ul),
        "in_list" / Int32ul,
        Ver("G >= G14 && V < V13_0B4", "unk_98_g14_0" / HexDump(Bytes(0x14))),
        "list_head" / LinkedListHead,
        Ver("G >= G14 && V < V13_0B4", "unk_a8_g14_0" / ZPadding(4)),
        Ver("V >= V13_0B4", "unk_buf" / EventControlUnkBuf),
    )

    def __init__(self):
        super().__init__()
        self.unk_14 = 0
        self.unk_18 = 0
        self.unk_20 = 0
        self.vm_slot = 0
        self.has_ta = 0
        self.pstamp_ta = [0]*4
        self.has_3d = 0
        self.pstamp_3d = [0]*4
        self.has_cp = 0
        self.pstamp_cp = [0]*4
        self.in_list = 0
        self.unk_98_g14_0 = bytes(0x14)
        self.list_head = LinkedListHead()
        self.unk_buf = EventControlUnkBuf()

class WorkCommandCP(ConstructClass):
    """
    For compute

    Example:
    00000000  00000003 00000000 00000004 0c3d80c0 ffffffa0 00000000 00000000 00000000
    00000020  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    00000040  *
    00000060  00000000 00000000 00088000 00000015 00078000 00000015 000a6300 00000015
    00000080  000a6308 00000015 000a6310 00000015 000a6318 00000015 00000000 00000011
    000000a0  00008c60 00000000 00000041 00000000 000e8000 00000015 00000040 00000000
    000000c0  00000001 00000000 0000001c 00000000 00000000 00000000 00000000 00000000
    000000e0  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    00000100  *
    000001e0  00000000 00000000 0c311cc0 ffffffa0 00000240 00000000 00000000 00000000
    00000200  00000000 00000000 00000000 00000000 00000000 00000000 00088000 00000015
    00000220  00078024 00000015 00000000 00000000 00000000 00000000 00000000 00000000
    00000240  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    00000260  110022b3 00000000 ffffffff 00000500 00000015 00000000 00000000 00000000
    00000280  000c8014 ffffffa0 0c378014 ffffffa0 00003b00 00000005 00000000 00000000
    000002a0  120022b8 00000000 00000000 00000000 00029030 ffffffa0 00029038 ffffffa0
    000002c0  00000000 00000000 00000000 00000000 00000015 00000000 00000000 00000000
    000002e0  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    """

    subcon = Struct(
        "addr" / Tell,
        "magic" / Const(0x3, Hex(Int32ul)),
        Ver("V >= V13_0B4", "counter" / Int64ul),
        "unk_4" / Hex(Int32ul),
        "context_id" / Hex(Int32ul),
        "event_control_addr" / Hex(Int64ul),
        "event_control" / ROPointer(this.event_control_addr, EventControl),
        "unk_2c" / Int32ul,
        Ver("G >= G14X", "registers" / Array(128, RegisterDefinition)),
        Ver("G >= G14X", "unk_g14x" / Default(Array(64, Int32ul), [0]*64)),
        Ver("G < G14X", "unk_buf" / HexDump(Bytes(0x50))),
        Ver("G < G14X", "compute_info" / ComputeInfo),
        "registers_addr" / Int64ul,
        "register_count" / Int16ul,
        "registers_length" / Int16ul,
        "unk_pad" / HexDump(Bytes(0x24)),
        "microsequence_ptr" / Hex(Int64ul),
        "microsequence_size" / Hex(Int32ul),
        "microsequence" / ROPointer(this.microsequence_ptr, MicroSequence),
        "compute_info2" / ComputeInfo2,
        "encoder_params" / EncoderParams,
        "job_meta" / JobMeta,
        "ts1" / TimeStamp,
        "ts_pointers" / TimeStampPointers,
        "user_ts_pointers" / TimeStampPointers, # This is a guess, but it makes sense
        "client_sequence" / Int8ul,
        Ver("V >= V13_0B4", "unk_ts2" / TimeStamp),
        Ver("V >= V13_0B4", "unk_ts" / TimeStamp),
        Ver("V >= V13_0B4", "unk_2e1" / Default(HexDump(Bytes(0x1c)), bytes(0x1c))),
        Ver("V >= V13_0B4", "unk_flag" / Flag),
        Ver("V >= V13_0B4", "unk_pad" / Default(HexDump(Bytes(0x10)), bytes(0x10))),
        "pad_2d9" / Default(HexDump(Bytes(0x7)), bytes(0x7)),
    )

class WorkCommand0_UnkBuf(ConstructValueClass):
    subcon = HexDump(Bytes(0x18))

    def __init__(self):
        self.value = bytes(0x18)

class WorkCommand1_UnkBuf(ConstructValueClass):
    subcon = HexDump(Bytes(0x110))

    def __init__(self):
        self.value = bytes(0x110)

class WorkCommand1_UnkBuf2(ConstructClass):
    subcon = Struct(
        "unk_0" / Int64ul,
        "unk_8" / Int64ul,
        "unk_10" / Int64ul,
    )

class WorkCommand3D(ConstructClass):
    """
    For 3D

    Example: 0xfa00c095640
    00000000  00000001 00000004 00000000 0c2d5f00 ffffffa0 000002c0 0c3d80c0 ffffffa0
    00000020  0c3e0000 ffffffa0 0c3e0100 ffffffa0 0c3e09c0 ffffffa0 01cb0000 00000015
    00000040  00000088 00000000 00000001 0010000c 00000000 00000000 00000000 00000000
    00000060  3a8de3be 3abd2fa8 00000000 00000000 0000076c 00000000 0000a000 00000000
    00000080  ffff8002 00000000 00028044 00000000 00000088 00000000 005d0000 00000015
    000000a0  00758000 00000015 0000c000 00000000 00000640 000004b0 0257863f 00000000
    000000c0  00000000 00000000 00000154 00000000 011d0000 00000015 011d0000 00000015
    000000e0  0195c000 00000015 0195c000 00000015 00000000 00000000 00000000 00000000
    00000100  00000000 00000000 00000000 00000000 0193c000 00000015 00000000 00000000
    00000120  0193c000 00000015 00000000 00000000 01b64000 00000015 00000000 00000000
    00000140  01b64000 00000015 00000000 00000000 01cb0000 00000015 01cb4000 00000015
    00000160  c0000000 00000003 01cb4000 00000015 00010280 00000000 00a38000 00000015
    00000180  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    000001a0  00000000 00000000 00000000 00000000 00000000 00000011 00008c60 00000000
    000001c0  00000000 00000000 00000000 00000000 0000001c 00000000 00000000 00000000
    000001e0  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    00000200  *
    000003c0  00000012 00028084 00000000 00000000 3a8de3be 3abd2fa8 00000000 00000000
    000003e0  0010000c 00000000 00025031 00000004 3f800000 00000700 00000000 00000001
    """

    subcon = Struct(
        "addr" / Tell,
        "magic" / Const(0x1, Hex(Int32ul)),
        Ver("V >= V13_0B4", "counter" / Int64ul),
        "context_id" / Hex(Int32ul),
        "unk_8" / Hex(Int32ul),
        "microsequence_ptr" / Hex(Int64ul), # Command list
        "microsequence_size" / Hex(Int32ul),
        "microsequence" / ROPointer(this.microsequence_ptr, MicroSequence),
        "event_control_addr" / Hex(Int64ul),
        "event_control" / ROPointer(this.event_control_addr, EventControl),
        "buffer_mgr_addr" / Int64ul,
        "buffer_mgr" / ROPointer(this.buffer_mgr_addr, BufferManagerInfo),
        "buf_thing_addr" / Int64ul,
        "buf_thing" / ROPointer(this.buf_thing_addr, BufferThing),
        "unk_emptybuf_addr" / Hex(Int64ul),
        "tvb_tilemap" / Hex(Int64ul),
        "unk_40" / Hex(Int64ul),
        "unk_48" / Hex(Int32ul),
        "tile_blocks_y" / Hex(Int16ul), # * 4
        "tile_blocks_x" / Hex(Int16ul), # * 4
        "unk_50" / Hex(Int64ul),
        "unk_58" / Hex(Int64ul),
        "merge_upper_x" / Hex(Float32l),
        "merge_upper_y" / Hex(Float32l),
        "unk_68" / Hex(Int64ul),
        "tile_count" / Hex(Int64ul),
        # Embedded structures that are also pointed to by other stuff
        Ver("G < G14X", "struct_2" / Start3DStruct2),
        Ver("G < G14X", "struct_1" / Start3DStruct1),
        Ver("G >= G14X", "registers" / Array(128, RegisterDefinition)),
        Ver("G >= G14X", "unk_g14x" / Default(Array(64, Int32ul), [0]*64)),
        "struct_3" / Start3DStruct3,
        "unk_758" / Flag,
        "unk_75c" / Flag,
        "unk_buf" / WorkCommand1_UnkBuf,
        "busy_flag" / Flag,
        "struct_6" / Start3DStruct6,
        "struct_7" / Start3DStruct7,
        "unk_buf2" / WorkCommand1_UnkBuf2,
        "ts1" / TimeStamp,
        "ts_pointers" / TimeStampPointers,
        "user_ts_pointers" / TimeStampPointers, # This is a guess, but it makes sense
        "client_sequence" / Int8ul,
        Ver("V >= V13_0B4", "unk_ts2" / TimeStamp),
        Ver("V >= V13_0B4", "unk_ts" / TimeStamp),
        Ver("V >= V13_0B4", "unk_pad3" / Default(HexDump(Bytes(0x20)), bytes(0x20))),
        "pad_925" / Default(HexDump(Bytes(0x3)), bytes(0x3)),
        Ver("V == V13_3", "unk_pad2" / Default(HexDump(Bytes(0x3c)), bytes(0x3c))),
    )

class WorkCommand0_UnkBuf(ConstructValueClass):
    subcon = HexDump(Bytes(0x18))

    def __init__(self):
        super().__init__()
        self.value = bytes(0x18)

class WorkCommandTA(ConstructClass):
    """
    For TA

    Example:
    00000000  00000000 00000004 00000000 0c3d80c0 ffffffa0 00000002 00000000 0c3e0000
    00000020  ffffffa0 0c3e0100 ffffffa0 0c3e09c0 ffffffa0 00000000 00000200 00000000
    00000040  1e3ce508 1e3ce508 01cb0000 00000015 00000000 00000000 00970000 00000015
    00000060  01cb4000 80000015 006b0003 003a0012 00000001 00000000 00000000 00000000
    00000080  0000a000 00000000 00000088 00000000 01cb4000 00000015 00000000 00000000
    000000a0  0000ff00 00000000 007297a0 00000015 00728120 00000015 00000001 00000000
    000000c0  00728000 00040015 009f8000 00000015 00000000 00000000 00000000 00000000
    000000e0  0000a441 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    00000100  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000011
    00000120  00000000 00000000 0000001c 00000000 00008c60 00000000 00000000 00000000
    00000140  00000000 00000000 00000000 00000000 0000001c 00000000 00000000 00000000
    00000160  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
    00000180  *
    000003a0  00000000 00000000 00000000 00000000 00000000 00000000 00000000 000000f0
    000003c0  00000088 00000202 04af063f 00025031 00404030 00303024 000000c0 00000180
    000003e0  00000100 00008000 00000000 00000000 00000000 00000000 00000000 00000000
    """

    subcon = Struct(
        "addr" / Tell,
        "magic" / Const(0x0, Hex(Int32ul)),
        Ver("V >= V13_0B4", "counter" / Int64ul),
        "context_id" / Hex(Int32ul),
        "unk_8" / Hex(Int32ul),
        "event_control_addr" / Hex(Int64ul),
        "event_control" / ROPointer(this.event_control_addr, EventControl),
        "buffer_mgr_slot" / Hex(Int64ul),
        "buffer_mgr_addr" / Int64ul,
        "buffer_mgr" / ROPointer(this.buffer_mgr_addr, BufferManagerInfo),
        "buf_thing_addr" / Int64ul,
        "buf_thing" / ROPointer(this.buf_thing_addr, BufferThing),
        "unk_emptybuf_addr" / Hex(Int64ul),
        "unk_34" / Hex(Int32ul),

        # Embedded structures that are also pointed to by other stuff
        Ver("G >= G14X", "registers" / Array(128, RegisterDefinition)),
        Ver("G >= G14X", "unk_154" / Default(HexDump(Bytes(0x100)), bytes(0x100))), # unknown
        Ver("G < G14X", "struct_2" / StartTACmdStruct2), # 0x11c bytes
        Ver("G < G14X", "unk_154" / Default(HexDump(Bytes(0x268)), bytes(0x268))), # unknown
        Ver("G < G14X", "tiling_params" / TilingParameters), # unknown
        Ver("G < G14X", "unk_3e8" / HexDump(Bytes(0x60))), # unknown
        "registers_addr" / Int64ul,
        "register_count" / Int16ul,
        "registers_length" / Int16ul,
        "unk_pad" / Int32ul,
        "unkptr_45c" / Int64ul,
        "tvb_size" / Int64ul,
        "microsequence_ptr" / Hex(Int64ul),
        "microsequence_size" / Hex(Int32ul),
        "microsequence" / ROPointer(this.microsequence_ptr, MicroSequence),
        "ev_3d" / Int32ul,
        "stamp_value" / Int32ul,

        "struct_3" / StartTACmdStruct3, # 0x114 bytes

        "unk_594" / WorkCommand0_UnkBuf,

        "ts1" / TimeStamp,
        "ts_pointers" / TimeStampPointers,
        "user_ts_pointers" / TimeStampPointers, # This is a guess, but it makes sense

        "client_sequence" / Int8ul,
        Ver("V >= V13_0B4", "unk_ts2" / TimeStamp),
        Ver("V >= V13_0B4", "unk_ts" / TimeStamp),
        Ver("V >= V13_0B4", "unk_5d8_15" / Default(HexDump(Bytes(0x18)), bytes(0x18))),
        "pad_5d5" / Default(HexDump(Bytes(0x3)), bytes(0x3)),
        "pad_5d8" / Default(HexDump(Bytes(0x8)), bytes(0x8)),
        Ver("V >= V13_3", "unk_pad2" / Default(HexDump(Bytes(0xc)), bytes(0xc))),
    )

class WorkCommandBlit(ConstructClass):
    subcon = Struct(
        "addr" / Tell,
        "magic" / Const(0x2, Int32ul),
        Ver("V >= V13_0B4", "counter" / Int64ul),
        "context_id" / Int32ul,
        "event_control_addr" / Hex(Int64ul),
        "event_control" / ROPointer(this.event_control_addr, EventControl),
        "unk_10" / Hex(Int32ul),
        "unk_14" / Int32ul,
        Ver("G < G14X", "blit_info" / BlitInfo),
        Ver("G >= G14X", "registers" / Array(128, RegisterDefinition)),
        Ver("G >= G14X", "unk_g14x" / Default(Array(64, Int32ul), [0]*64)),
        "registers_addr" / Int64ul,
        "register_count" / Int16ul,
        "registers_length" / Int16ul,
        "blit_info2" / BlitInfo2,
        "microsequence_ptr" / Hex(Int64ul),
        "microsequence_size" / Hex(Int32ul),
        "microsequence" / ROPointer(this.microsequence_ptr, MicroSequence),
        "unkptr_49c" / HexDump(Bytes(0x114)),
        "job_meta" / JobMeta,
        "unk_5d8" / Int32ul,
        "unk_stamp_ptr" / Int64ul,
        "unk_stamp_val" / Int32ul,
        "unk_5ec" / Int32ul,
        "encoder_params" / EncoderParams,
        "unk_618" / Int32ul,
        "ts1" / TimeStamp,
        "ts_pointers" / TimeStampPointers,
        "user_ts_pointers" / TimeStampPointers,
        "client_sequence" / Int8ul,
        "pad_645" / Default(HexDump(Bytes(0x3)), bytes(0x3)),
        "unk_648" / Int32ul,
        "unk_64c" / Int8ul,
        "pad_64d" / Default(HexDump(Bytes(0x7)), bytes(0x7)),
    )

class UnknownWorkCommand(ConstructClass):
    subcon = Struct(
        "magic" / Hex(Int32ul),
        "unk_4" / Hex(Int32ul),
        "unk_8" / Hex(Int32ul),
        "unk_c" / Hex(Int32ul),
        "unk_10" / Hex(Int32ul),
        "unk_14" / Hex(Int32ul),
        "unk_18" / Hex(Int32ul),
        "unk_1c" / Hex(Int32ul),
    )

class CmdBufWork(ConstructClass):
    subcon = Struct(
        "cmdid" / Peek(Int32ul),
        "cmd" / Switch(this.cmdid, {
            0: WorkCommandTA,
            1: WorkCommand3D,
            2: WorkCommandBlit,
            3: WorkCommandCP,
            4: WorkCommandBarrier,
            6: WorkCommandInitBM,
            10: WorkCommandComputeUnk10,
            11: WorkCommandComputeUnk11,
        })
    )

class JobList(ConstructClass):
    subcon = Struct(
        "first_job" / Default(Int64ul, 0),
        "last_head" / Int64ul,
        "unkptr_10" / Default(Int64ul, 0),
    )

class GPUContextData(ConstructClass):
    subcon = Struct(
        "queue_table_index" / Int8ul,
        "pid_table_index" / Int8ul,
        "unk_2" / Default(Bytes(3), bytes(3)),
        "unk_5" / Int8ul,
        "unk_6" / Int32ul,
        "unk_a" / Int32ul,
        "unk_e" / Int32ul,
        "unk_12" / Int32ul,
        "unk_16" / Int32ul,
        "unk_1a" / Int32ul,
        "unk_1e" / Int8ul,
        "unk_1f" / Int8ul,
        "unk_20" / Default(Bytes(3), bytes(3)),
        "unk_23" / Int8ul,
        "unk_24" / Default(Bytes(0x1c), bytes(0x1c)),
    )

    def __init__(self):
        self.queue_table_index = 0xff
        self.pid_table_index = 0xff
        self.unk_5 = 1
        self.unk_6 = 0
        self.unk_a = 0
        self.unk_e = 0
        self.unk_12 = 0
        self.unk_16 = 0
        self.unk_1a = 0
        self.unk_1e = 0xff
        self.unk_1f = 0
        self.unk_23 = 2

class CommandQueuePointerMap(RegMap):
    GPU_DONEPTR = 0x00, Register32
    GPU_RPTR = 0x30, Register32
    CPU_WPTR = 0x40, Register32

class CommandQueuePointers(ConstructClass):
    subcon = Struct(
        "gpu_doneptr" / Int32ul,
        ZPadding(12),
        "unk_10" / Int32ul,
        ZPadding(12),
        "unk_20" / Int32ul,
        ZPadding(12),
        "gpu_rptr" / Int32ul,
        ZPadding(12),
        "cpu_wptr" / Int32ul,
        ZPadding(12),
        "rb_size" / Int32ul,
        ZPadding(12),
    )

    def __init__(self):
        super().__init__()
        self.gpu_doneptr = 0
        self.unk_10 = 0
        self.unk_20 = 0
        self.gpu_rptr = 0
        self.cpu_wptr = 0
        self.rb_size = 0x500

class CommandQueueInfo(ConstructClass):
    """ Structure type shared by Submit3D, SubmitTA and SubmitCompute
        Applications have multiple of these, one of each submit type
        TODO: Can applications have more than one of each type? One per encoder?
        Mostly managed by GPU, only initialize by CPU

    """
    subcon = Struct(
        "pointers_addr" / Hex(Int64ul),
        "pointers" / ROPointer(this.pointers_addr, CommandQueuePointers),
        "rb_addr" / Hex(Int64ul), # 0x4ff pointers
        "job_list_addr" / Hex(Int64ul), # ffffffa000000000, size 0x18 (shared by 3D and TA)
        "job_list" / ROPointer(this.job_list_addr, JobList),
        "gpu_buf_addr" / Hex(Int64ul), # GPU space for this queue, 0x2c18 bytes?
        #"gpu_buf" / ROPointer(this.gpu_buf_addr, HexDump(Bytes(0x2c18))),
        "gpu_rptr1" / Hex(Int32ul),
        "gpu_rptr2" / Hex(Int32ul),
        "gpu_rptr3" / Hex(Int32ul),
        "event_id" / Int32sl,
        "priority" / Hex(Int32ul), # read by CPU
        "unk_34" / Hex(Int32ul),
        "unk_38" / Hex(Int64ul),
        "unk_40" / Hex(Int32ul), # 1
        "unk_44" / Hex(Int32ul), # 0
        "prio5" / Hex(Int32ul), # 1, 2
        "unk_4c" / Int32sl, # -1
        "uuid" / Hex(Int32ul), # Counts up for each new process or command queue
        "unk_54" / Int32sl,
        "unk_58" / Hex(Int64ul), # 0
        "busy" / Hex(Int32ul), # 1 = gpu busy
        "pad1" / ZPadding(0x1c),
        "unk_80" / Hex(Int32ul),
        "has_commands" / Hex(Int32ul),
        "unk_88" / Int32ul,
        "unk_8c" / Int32ul,
        "unk_90" / Int32ul,
        "unk_94" / Int32ul,
        "inflight_commands" / Int32ul,
        "unk_9c" / Int32ul,
        Ver("V >= V13_2 && G < G14X", "unk_a0_0" / Int32ul),
        "gpu_context_addr" / Hex(Int64ul), # GPU managed context, shared between 3D and TA. Passed to DC_DestroyContext
        "gpu_context" / ROPointer(this.gpu_context_addr, GPUContextData),
        "unk_a8" / Int64ul,
        Ver("V >= V13_2 && G < G14X", "unk_b0" / Int32ul),
        # End of struct
    )

    def __init__(self):
        super().__init__()
        self.gpu_rptr1 = 0
        self.gpu_rptr2 = 0
        self.gpu_rptr3 = 0
        self.event_id = -1
        self.unk_4c = -1
        self.uuid = 0xdeadbeef # some kind of ID
        self.unk_54 = -1
        self.unk_58 = 0x0
        self.busy = 0x0
        self.unk_80 = 0
        self.has_commands = 0
        self.unk_88 = 0
        self.unk_8c = 0
        self.unk_90 = 0
        self.unk_94 = 0
        self.inflight_commands = 0
        self.unk_9c = 0
        self.unk_a0_0 = 0
        self.set_prio(0)
        self.unk_a8 = 0
        self.unk_b0 = 0

    def set_prio(self, p):
        if p == 0:
            self.priority = 0
            self.unk_34 = 0 # 0-3?
            self.unk_38 = 0xffff_ffff_ffff_0000
            self.unk_40 = 1
            self.unk_44 = 0
            self.prio5 = 1
        elif p == 1:
            self.priority = 1
            self.unk_34 = 1
            self.unk_38 = 0xffff_ffff_0000_0000
            self.unk_40 = 0
            self.unk_44 = 0
            self.prio5 = 0
        elif p == 2:
            self.priority = 2
            self.unk_34 = 2
            self.unk_38 = 0xffff_0000_0000_0000
            self.unk_40 = 0
            self.unk_44 = 0
            self.prio5 = 2
        else:
            self.priority = 3
            self.unk_34 = 3
            self.unk_38 = 0x0000_0000_0000_0000
            self.unk_40 = 0
            self.unk_44 = 0
            self.prio5 = 3

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

# speaker_amp.py -- play audio through the embedded speaker on Mac mini
#
# sample usage with sox:
#
#   sox INPUT_FILE -t raw -r 48000 -c 1 -e signed-int -b 32 -L - gain -63 | python3 ./speaker_amp.py
#
# (expects mono, 24-bit signed samples padded to 32 bits on the msb side)

import argparse
from m1n1.setup import *
from m1n1.hw.dart import DART, DARTRegs
from m1n1.hw.i2c import I2C
from m1n1.hw.pmgr import PMGR
from m1n1.hw.nco import NCO
from m1n1.hw.admac import *
from m1n1.hw.mca import *

argparser = argparse.ArgumentParser()
argparser.add_argument("-f", "--file", "--input", "--samples",
                       type=str, default=None,
                       help='input filename to take samples from ' \
                            '(default: standard input)')
argparser.add_argument("-b", "--bufsize", type=int, default=1024*32,
                       help='size of buffers to keep submitting to DMA')
args = argparser.parse_args()

inputf = open(args.file, "rb") if args.file is not None \
            else sys.stdin.buffer 

p.pmgr_adt_clocks_enable("/arm-io/i2c1")
p.pmgr_adt_clocks_enable("/arm-io/admac-sio")
p.pmgr_adt_clocks_enable("/arm-io/dart-sio")
p.pmgr_adt_clocks_enable("/arm-io/mca-switch")

# reset AUDIO_P
PS_AUDIO_P = PMGR(u).regs[0].PS4[5]
PS_AUDIO_P.set(DEV_DISABLE=1)
PS_AUDIO_P.set(RESET=1)
PS_AUDIO_P.set(RESET=0)
PS_AUDIO_P.set(DEV_DISABLE=0)

i2c1 = I2C(u, "/arm-io/i2c1")

dart_base, _ = u.adt["/arm-io/dart-sio"].get_reg(0) # stream index 2
dart = DART(iface, DARTRegs(u, dart_base), util=u)
dart.initialize()

cl_no = 0

admac = ADMAC(u, "/arm-io/admac-sio", dart, debug=True)
tx_chan = admac.chans[4*cl_no]

tx_chan.disable()
tx_chan.reset()
tx_chan.read_reports() # read stale reports
tx_chan.buswidth = E_BUSWIDTH.W_32BIT
tx_chan.framesize = E_FRAME.F_1_WORD

nco = NCO(u, "/arm-io/nco")
nco[cl_no].set_rate(48000 * 256)
nco[cl_no].enable()

mca_switch1_base = u.adt["/arm-io/mca-switch"].get_reg(1)[0]
mca_cl_base = u.adt["/arm-io/mca-switch"].get_reg(0)[0] + 0x4000*cl_no
cl = MCACluster(u, mca_cl_base)

regs, serdes = cl.regs, cl.txa

regs.SYNCGEN_STATUS.set(RST=1, EN=0)
regs.SYNCGEN_STATUS.set(RST=0)
regs.SYNCGEN_MCLK_SEL.val =(1 + cl_no)
regs.SYNCGEN_HI_PERIOD.val = 0
regs.SYNCGEN_LO_PERIOD.val = 0xfe  # full period minus two

serdes.STATUS.set(EN=0)
serdes.CONF.set(
    NSLOTS=0,
    SLOT_WIDTH=E_SLOT_WIDTH.W_32BIT,
    BCLK_POL=1,
    UNK1=1, UNK2=1,
    IDLE_UNDRIVEN=1,
    SYNC_SEL=(1 + cl_no)
)
serdes.BITDELAY.val = 0

serdes.CHANMASK[0].val = 0xffff_fffe
serdes.CHANMASK[1].val = 0xffff_fffe

regs.PORT_ENABLES.set(CLOCK1=1, CLOCK2=1, DATA=1)
regs.PORT_CLK_SEL.set(SEL=(cl_no + 1))
regs.PORT_DATA_SEL.val = cl_no + 1
regs.MCLK_STATUS.set(EN=1)
regs.SYNCGEN_STATUS.set(EN=1)

p.write32(mca_switch1_base + 0x8000*cl_no, 0x102048)

# toggle the GPIO line driving the speaker-amp IC reset
p.write32(0x23c1002d4, 0x76a02) # invoke reset
p.write32(0x23c1002d4, 0x76a03) # take out of reset

tx_chan.submit(inputf.read(args.bufsize))
tx_chan.enable()
while tx_chan.can_submit():
    tx_chan.submit(inputf.read(args.bufsize))

serdes.STATUS.set(EN=1)

# by ADT and leaked schematic, i2c1 contains TAS5770L,
# which is not a public part. but there's e.g. TAS2770
# with similar registers
#
# https://www.ti.com/product/TAS2770
#
# if the speaker-amp IC loses clock on the serial sample input,
# it automatically switches to software shutdown.
#

i2c1.write_reg(0x31, 0x08, [0x40])
i2c1.write_reg(0x31, 0x0a, [0x06, 0x00, 0x1a])
i2c1.write_reg(0x31, 0x1b, [0x01, 0x82, 0x06])
i2c1.write_reg(0x31, 0x16, [0x50, 0x04])
i2c1.write_reg(0x31, 0x0d, [0x00])
#i2c1.write_reg(0x31, 0x03, [0x14])

# amplifier gain, presumably this is the lowest setting
i2c1.write_reg(0x31, 0x03, [0x0])

# take the IC out of software shutdown
i2c1.write_reg(0x31, 0x02, [0x0c])

while (buf := inputf.read(args.bufsize)):
    while not tx_chan.can_submit():
        tx_chan.poll()
    tx_chan.submit(buf)

# mute
i2c1.write_reg(0x31, 0x02, [0x0d])

# software shutdown
i2c1.write_reg(0x31, 0x02, [0x0e])

tx_chan.disable()



# SPDX-License-Identifier: MIT

from m1n1.trace.i2c import I2CTracer

I2CTracer = I2CTracer._reloadcls()

i2c_tracers = {}

for node in hv.adt["/arm-io"]:
    if node.name.startswith("i2c"):
        n = int(node.name[3:])
        i2c_tracers[n] = I2CTracer(hv, f"/arm-io/i2c{n}", verbose=0)
        i2c_tracers[n].stop()
        i2c_tracers[n].start()
        if hv.ctx:
            for irq in getattr(node, "interrupts"):
                hv.trace_irq(node.name, irq, 1, hv.IRQTRACE_IRQ)

from m1n1.gpiola import GPIOLogicAnalyzer

if not hv.started:
    for cpu in list(hv.adt["cpus"]):
        if cpu.name == "cpu3":
            print(f"Removing ADT node {cpu._path}")
            del hv.adt["cpus"][cpu.name]

if not hv.started or hv.ctx is not None:
    m = GPIOLogicAnalyzer(u, "arm-io/gpio",
                        pins={"scl": 0xc9, "sda": 0xc7},
                        div=1, on_pin_change=True, cpu=3)

    m.load_regmap(list(i2c_tracers[1].regmaps.values())[0],
                  regs={"SMSTA", "XFSTA"})

def start_la():
    m.start(1000000, bufsize=0x80000)
    hv.cont()

def stop_la():
    m.complete()
    m.show()



# SPDX-License-Identifier: MIT
from ...utils import *

from .crash import ASCCrashLogEndpoint
from .syslog import ASCSysLogEndpoint
from .mgmt import ASCManagementEndpoint
from .kdebug import ASCKDebugEndpoint
from .ioreporting import ASCIOReportingEndpoint
from .oslog import ASCOSLogEndpoint
from .base import ASCBaseEndpoint, ASCTimeout
from ...hw.asc import ASC

__all__ = []

class ASCDummyEndpoint(ASCBaseEndpoint):
    SHORT = "dummy"

class StandardASC(ASC):
    ENDPOINTS = {
        0: ASCManagementEndpoint,
        1: ASCCrashLogEndpoint,
        2: ASCSysLogEndpoint,
        3: ASCKDebugEndpoint,
        4: ASCIOReportingEndpoint,
        8: ASCOSLogEndpoint,
        0xa: ASCDummyEndpoint, # tracekit
    }

    def __init__(self, u, asc_base, dart=None, stream=0):
        super().__init__(u, asc_base)
        self.remote_eps = set()
        self.add_ep(0, ASCManagementEndpoint(self, 0))
        self.dart = dart
        self.stream = stream
        self.eps = []
        self.epcls = {}
        self.dva_offset = 0
        self.dva_size = 1 << 32
        self.allow_phys = False

        for cls in type(self).mro():
            eps = getattr(cls, "ENDPOINTS", None)
            if eps is None:
                break
            for k, v in eps.items():
                if k not in self.epcls:
                    self.epcls[k] = v

    def addr(self, addr):
        return f"{addr:#x}"

    def iomap(self, addr, size):
        if self.dart is None:
            return addr
        dva = self.dva_offset | self.dart.iomap(self.stream, addr, size)

        self.dart.invalidate_streams(1)
        return dva

    def ioalloc(self, size):
        paddr = self.u.memalign(0x4000, size)
        dva = self.iomap(paddr, size)
        return paddr, dva

    def ioread(self, dva, size):
        if self.allow_phys and dva < self.dva_offset or dva >= (self.dva_offset + self.dva_size):
            return self.iface.readmem(dva, size)

        if self.dart:
            return self.dart.ioread(self.stream, dva & 0xFFFFFFFFF, size)
        else:
            return self.iface.readmem(dva, size)

    def iowrite(self, dva, data):
        if self.allow_phys and dva < self.dva_offset or dva >= (self.dva_offset + self.dva_size):
            return self.iface.writemem(dva, data)

        if self.dart:
            return self.dart.iowrite(self.stream, dva & 0xFFFFFFFFF, data)
        else:
            return self.iface.writemem(dva, data)

    def iotranslate(self, dva, size):
        if self.allow_phys and dva < self.dva_offset or dva >= (self.dva_offset + self.dva_size):
            return [(dva, size)]

        if self.dart:
            return self.dart.iotranslate(self.stream, dva & 0xFFFFFFFFF, size)
        else:
            return [(dva, size)]

    def start_ep(self, epno):
        if epno not in self.epcls:
            raise Exception(f"Unknown endpoint {epno:#x}")

        epcls = self.epcls[epno]
        ep = epcls(self, epno)
        self.add_ep(epno, ep)
        print(f"Starting endpoint #{epno:#x} ({ep.name})")
        self.mgmt.start_ep(epno)
        ep.start()

    def start(self):
        super().boot()
        self.mgmt.start()
        self.mgmt.wait_boot(3)

    def stop(self, state=0x10):
        for ep in list(self.epmap.values())[::-1]:
            if ep.epnum < 0x10:
                continue
            ep.stop()
        self.mgmt.stop(state=state)
        self.epmap = {}
        self.add_ep(0, ASCManagementEndpoint(self, 0))
        if state < 0x10:
            self.shutdown()

    def boot(self):
        print("Booting ASC...")
        super().boot()
        self.mgmt.wait_boot(1)

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



# SPDX-License-Identifier: MIT
import io, time

from ..malloc import Heap
from ..utils import *
from ..constructutils import ConstructClassBase, str_value
from construct import Bytes, Container, HexDump
from ..hw.uat import MemoryAttr

class GPUObject:
    def __init__(self, allocator, objtype):
        self._raw = False
        if isinstance(objtype, int):
            self.val = bytes(objtype)
            self._size = objtype
            self._name = b"Bytes({objtype})"
            self._raw = True
        elif isinstance(objtype, ConstructClassBase):
            self.val = objtype
            objtype = type(objtype)
            self._size = objtype.sizeof()
            self._name = objtype.__name__
        elif isinstance(objtype, type) and issubclass(objtype, ConstructClassBase):
            self._size = objtype.sizeof()
            self.val = objtype()
            self._name = objtype.__name__
        else:
            self._size = objtype.sizeof()
            self.val = objtype.parse(bytes(self._size))
            self._name = type(objtype).__name__

        self._alloc = allocator
        self._type = objtype
        self._addr = None
        self._data = None
        self._dead = False
        self._map_flags = {}
        self._mon_val = None
        self._skipped_pushes = 0
        self._compress_threshold = 65536
        self._strm = None
        self._read_phys = False

    def push(self, if_needed=False):
        self._mon_val = self.val
        assert self._addr is not None

        if self._raw:
            data = self.val
        else:
            context = Container()
            context._parsing = False
            context._building = True
            context._sizing = False
            context._params = context
            # build locally and push as a block for efficiency
            ios = io.BytesIO()
            self._type._build(self.val, ios, context, "(pushing)")
            data = ios.getvalue()

        #if self._alloc.verbose:
            #t = time.time()
            #self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] chk {self._size} bytes")
        if if_needed and data[:] == self._data:
            self._skipped_pushes += 1
            #if self._alloc.verbose:
                #t2 = time.time()
                #mbs = self._size / (t2 - t) / 1000000
                #self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] chk done ({mbs:.02f} MB/s)")
            return self

        self._skipped_pushes = 0

        t = time.time()
        if data == bytes(self._size):
            if self._alloc.verbose:
                self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] zeroing {self._size} bytes")
            self._alloc.agx.p.memset8(self._paddr, 0, self._size)
        elif self._size > self._compress_threshold:
            if self._alloc.verbose:
                self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] pushing {self._size} bytes (compressed)")
            self._alloc.agx.u.compressed_writemem(self._paddr, data)
        else:
            if self._alloc.verbose:
                self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] pushing {self._size} bytes")
            self._alloc.agx.iface.writemem(self._paddr, data)
        if self._alloc.verbose:
            t2 = time.time()
            mbs = self._size / (t2 - t) / 1000000
            self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] push done ({mbs:.02f} MB/s)")
        #stream.write(data)
        if isinstance(self._type, type) and issubclass(self._type, ConstructClassBase):
            if self._strm is None:
                self._strm = self._alloc.make_stream(self._addr)
            self.val.set_addr(self._addr, self._strm)

        self._data = bytes(data)
        return self

    def _pull(self):
        if self._raw:
            assert self._paddr is not None
            return self._alloc.agx.iface.readmem(self._paddr, self._size)

        assert self._addr is not None
        context = Container()
        context._parsing = True
        context._building = False
        context._sizing = False
        context._params = context
        if self._alloc.verbose:
            self._alloc.agx.log(f"[{self._name} @{self._addr:#x}] pulling {self._size} bytes")
        if self._read_phys:
            stream = io.BytesIO()
            stream.write(self._alloc.agx.iface.readmem(self._paddr, self._size))
            stream.seek(0)
        else:
            stream = self._alloc.make_stream(self._addr)
        return self._type._parse(stream, context, f"(pulling {self._name})")

    def pull(self):
        self._mon_val = self.val = self._pull()
        return self

    def poll(self):
        prev_val = self._mon_val
        self._mon_val = cur_val = self._pull()
        if not hasattr(cur_val, "diff"):
            return None
        if cur_val != prev_val:
            diff = cur_val.diff(prev_val)
            assert diff is not None
            return f"GPUObject {self._name} ({self._size:#x} @ {self._addr:#x}): " + diff
        else:
            return None

    @property
    def _ctx(self):
        return self._alloc.ctx

    def add_to_mon(self, mon):
        mon.add(self._addr, self._size, self._name, offset=0,
                readfn=lambda a, s: self._alloc.agx.iface.readmem(a - self._addr + self._paddr, s))

    def _set_addr(self, addr, paddr=None):
        self._addr = addr
        self._paddr = paddr
        if isinstance(self.val, ConstructClassBase):
            self.val.set_addr(addr)

    def __getitem__(self, item):
        return self.val[item]
    def __setitem__(self, item, value):
        self.val[item] = value

    def __getattr__(self, attr):
        return getattr(self.val, attr)

    def __setattr__(self, attr, val):
        if attr.startswith("_") or attr == "val":
            self.__dict__[attr] = val
            return

        setattr(self.val, attr, val)

    def __str__(self):
        if isinstance(self.val, bytes) and len(self.val) > 128:
            s_val = f"<{len(self.val)} bytes>"
        else:
            s_val = str_value(self.val)
        return f"GPUObject {self._name} ({self._size:#x} @ {self._addr:#x}): " + s_val

    def free(self):
        if self._dead:
            return
        self._dead = True
        self._alloc.free(self)

class GPUAllocator:
    def __init__(self, agx, name, start, size,
                 ctx=0, page_size=16384, va_block=None, guard_pages=1, **kwargs):
        self.page_size = page_size
        if va_block is None:
            va_block = page_size
        self.agx = agx
        self.ctx = ctx
        self.name = name
        self.start = start
        self.size = size
        self.va = Heap(start, start + size, block=va_block)
        self.verbose = 0
        self.guard_pages = guard_pages
        self.objects = {}
        self.flags = kwargs
        self.align_to_end = True

    def make_stream(self, base):
        return self.agx.uat.iostream(self.ctx, base, recurse=False)

    def new(self, objtype, name=None, track=True, align=1, **kwargs):
        obj = GPUObject(self, objtype)
        obj._stream = self.make_stream
        if name is not None:
            obj._name = name

        guard_size = self.page_size * self.guard_pages

        size_align = align_up(obj._size, self.page_size)
        addr = self.va.malloc(size_align + guard_size)
        paddr = self.agx.u.memalign(self.page_size, size_align)
        off = 0
        if self.align_to_end:
            off = size_align - align_up(obj._size, align)

        flags = dict(self.flags)
        flags.update(kwargs)

        obj._addr_align = addr
        obj._paddr_align = paddr
        obj._size_align = size_align
        self.agx.uat.iomap_at(self.ctx, addr, paddr, size_align, **flags)
        obj._set_addr(addr + off, paddr + off)
        obj._map_flags = flags

        self.objects[obj._addr] = obj

        if self.verbose:
            self.agx.log(f"[{self.name}] Alloc {obj._name} size {obj._size:#x} @ {obj._addr:#x} ({obj._paddr:#x})")

        self.agx.reg_object(obj, track=track)
        return obj

    def new_buf(self, size, name, track=True):
        return self.new(HexDump(Bytes(size)), name=name, track=track)

    def buf(self, size, name, track=True):
        return self.new_buf(size, name, track).push()._addr

    def free(self, obj):
        obj._dead = True
        is_private = obj._map_flags.get("AttrIndex", MemoryAttr.Normal) != MemoryAttr.Shared
        if is_private and obj._addr_align > 0xf8000000000:
            flags2 = dict(obj._map_flags)
            flags2["AttrIndex"] = MemoryAttr.Shared
            self.agx.uat.iomap_at(self.ctx, obj._addr_align, obj._paddr_align,
                                  obj._size_align, **flags2)
            self.agx.uat.flush_dirty()
            self.agx.uat.handoff.prepare_cacheflush(obj._addr_align, obj._size_align)
            self.agx.ch.fwctl.send_inval(0x40, obj._addr_align)
            self.agx.uat.handoff.wait_cacheflush()

        self.agx.uat.iomap_at(self.ctx, obj._addr_align, 0,
                              obj._size_align, VALID=0)

        if is_private and obj._addr_align > 0xf8000000000:
            self.agx.uat.flush_dirty()
            self.agx.uat.handoff.complete_cacheflush()

        self.agx.u.free(obj._paddr_align)
        self.va.free(obj._addr_align)
        del self.objects[obj._addr]
        self.agx.unreg_object(obj)

        if self.verbose:
            self.agx.log(f"[{self.name}] Free {obj._name} size {obj._size:#x} @ {obj._addr:#x} ({obj._paddr:#x})")



# SPDX-License-Identifier: MIT
from ..utils import *
from collections import namedtuple
from enum import IntEnum


EncodeNotRawDescriptor = namedtuple('EncodeNotRawDescriptor', [
    'flags',                                # +0x000
        # [31:16] ?????
        # 13 ?????
        # 12 ?????
        # [11:9] ?????
        # 8 - enable alpha
        # 7 - alpha channel bpp
        #   0 -> 8bpp
        #   1 -> 16bpp
        # 6 - something unknown about tiling
        # [5:4] ?????
        # [3:2] - chroma subsampling
        #   00 -> broken?
        #   01 -> broken?
        #   10 -> 4:2:2
        #   11 -> 4:4:4
        # [1:0] - input bpp
        #   00 -> 8bpp
        #   01 -> 16bpp?
        #   10 -> 16bpp?
        #   11 -> 16bpp?
        #       the last three all produce slightly different outputs
        #       so might be 10/12/14/16?????
    'flags2',                               # +0x004
    'output_iova',                          # +0x008
    'max_out_sz',                           # +0x010
    'offset_x',                             # +0x014
    'offset_y',                             # +0x016
    'pix_surface_w_2_',                     # +0x018
    'pix_surface_h_2_',                     # +0x01a
    'pix_surface_w',                        # +0x01c
    'pix_surface_h',                        # +0x01e
    'luma_stride',                          # +0x020
    'chroma_stride',                        # +0x022
    'alpha_stride',                         # +0x024
    'unk_pad_0x26_',                        # +0x026
    'luma_iova',                            # +0x028
    'pix_plane0_tileheader_thing_',         # +0x030
    'chroma_iova',                          # +0x038
    'pix_plane1_tileheader_thing_',         # +0x040
    'alpha_iova',                           # +0x048
    'pix_plane2_tileheader_thing_',         # +0x050
    'frame_header_sz',                      # +0x058
    'unk_pad_0x5a_',                        # +0x05a
    'bitstream_version',                    # +0x05b
    'encoder_identifier',                   # +0x05c
    'pix_surface_w_byteswap_',              # +0x060
    'pix_surface_h_byteswap_',              # +0x062
    'chroma_format_interlace_mode',         # +0x064
    'aspect_ratio_frame_rate',              # +0x065
    'color_primaries',                      # +0x066
    'transfer_characteristic',              # +0x067
    'matrix_coefficients',                  # +0x068
    'alpha_channel_type',                   # +0x069
    'frame_hdr_reserved14',                 # +0x06a
    'unk_pad_0x6c_',                        # +0x06c
    'deprecated_number_of_slices',          # +0x0ec
    'log2_desired_slice_size_in_mb',        # +0x0ee
    'quantization_index',                   # +0x0ef
    'unk_0xf0_',                            # +0x0f0
    'unk_0xf2_',                            # +0x0f2
    'unk_0xf4_',                            # +0x0f4
    'unk_0xfc_',                            # +0x0fc
    'unk_0x100_0_',                         # +0x100
    'unk_0x100_1_',                         # +0x104
    'unk_0x100_2_',                         # +0x108
    'unk_0x100_3_',                         # +0x10c
    'unk_0x110_0_',                         # +0x110
    'unk_0x110_1_',                         # +0x114
    'unk_0x110_2_',                         # +0x118
    'unk_0x110_3_',                         # +0x11c
    'unk_0x110_4_',                         # +0x120
    'unk_0x110_5_',                         # +0x124
    'unk_0x110_6_',                         # +0x128
    'unk_0x110_7_',                         # +0x12c
    'unk_0x110_8_',                         # +0x130
    'unk_0x110_9_',                         # +0x134
    'unk_0x110_10_',                        # +0x138
    'unk_0x110_11_',                        # +0x13c
    'unk_0x110_12_',                        # +0x140
    'unk_0x110_13_',                        # +0x144
    'unk_0x110_14_',                        # +0x148
    'unk_0x110_15_',                        # +0x14c
    'quant_table_sel',                      # +0x150
        # upper nibble: quality / table index
        # lower nibble UNKNOWN!
    'unk_pad_0x154_',                       # +0x154
])
ENCODE_NOT_RAW_STRUCT = "<IIQIHHHHHHHHH2sQQQQQQH1sBIHHBBBBBB2s128sHBBHHQIIIIIIIIIIIIIIIIIIIIII44s"


class ProResRegs(RegMap):
    # something reads
    REG_0x0     = 0x000, Register32
    MODE        = 0x008, Register32     # 4 bits
    IRQ_ENABLE  = 0x00c, Register32     # 2 bits
    IRQ_STATUS  = 0x010, Register32

    ST0         = 0x014, Register32     # interrupt handler reads
    ST1         = 0x018, Register32     # interrupt handler reads
    REG_0x1c    = 0x01c, Register32     # interrupt handler reads
    REG_0x38    = 0x038, Register32     # exists, maybe RO
    REG_0x3c    = 0x03c, Register32     # interrupt handler reads
    REG_0x40    = 0x040, Register32     # exists, maybe RO, looks like 0x44
    REG_0x44    = 0x044, Register32     # interrupt handler reads
    REG_0x48    = 0x048, Register32     # exists, maybe RO, looks like 0x44
    REG_0x4c    = 0x04c, Register32     # exists, maybe RO, looks like 0x44
    REG_0x50    = 0x050, Register32     # exists, maybe RO, looks like 0x44
    REG_0x54    = 0x054, Register32     # exists, maybe RO, looks like 0x44

    DR_SIZE     = 0x100, Register32
    DR_ADDR_LO  = 0x104, Register32
    DR_ADDR_HI  = 0x108, Register32
    DR_HEAD     = 0x10c, Register32     # bit24 is special, something about wrapping around?
    DR_TAIL     = 0x110, Register32

    # This giant block may or may not be touched by tunables
    # Function is all unknown
    REG_0x114   = 0x114, Register32     # can set bits 0000FFFF
    REG_0x118   = 0x118, Register32     # can set bits 07FF07FF

    REG_0x134   = 0x134, Register32     # can set bits 00000003

    REG_0x144   = 0x144, Register32     # can set bits 00000001
    REG_0x148   = 0x148, Register32     # can set bits 00000001

    REG_0x160   = 0x160, Register32     # can set bits BFFF3FFF
    REG_0x164   = 0x164, Register32     # can set bits 07FF07FF

    REG_0x170   = 0x170, Register32     # can set bits BFFF3FFF
    REG_0x174   = 0x174, Register32     # can set bits 07FF07FF

    REG_0x180   = 0x180, Register32     # can set bits BFFF3FFF
    REG_0x184   = 0x184, Register32     # can set bits 07FF07FF

    REG_0x190   = 0x190, Register32     # can set bits BFFF3FFF
    REG_0x194   = 0x194, Register32     # can set bits 000000FF
    REG_0x198   = 0x198, Register32     # RO? init value 07FB066F

    REG_0x1a0   = 0x1a0, Register32     # can set bits BFFF3FFF
    REG_0x1a4   = 0x1a4, Register32     # can set bits 000000FF
    REG_0x1a8   = 0x1a8, Register32     # RO? init value 037C03EE

    REG_0x1b0   = 0x1b0, Register32     # can set bits BFFF3FFF
    REG_0x1b4   = 0x1b4, Register32     # can set bits 000000FF
    REG_0x1b8   = 0x1b8, Register32     # RO? init value 04E00377

    REG_0x1c0   = 0x1c0, Register32     # can set bits BFFF3FFF
    REG_0x1c4   = 0x1c4, Register32     # can set bits 000000FF
    REG_0x1c8   = 0x1c8, Register32     # RO? init value 051C00DA

    REG_0x1d0   = 0x1d0, Register32     # can set bits BFFF3FFF
    REG_0x1d4   = 0x1d4, Register32     # can set bits 000000FF
    REG_0x1d8   = 0x1d8, Register32     # can set bits 000000FF
    REG_0x1dc   = 0x1dc, Register32     # can set bits 00FFFFFF

    REG_0x1ec   = 0x1ec, Register32     # can set bits FFFFFFFF

    REG_0x270   = 0x270, Register32     # can set bits BFFF3FFF
    REG_0x274   = 0x274, Register32     # can set bits 07FF07FF
    REG_0x278   = 0x278, Register32     # can set bits FFFFFFC0
    REG_0x27c   = 0x27c, Register32     # can set bits 000003FF
    REG_0x280   = 0x280, Register32     # can set bits FFFFFFC0
    REG_0x284   = 0x284, Register32     # can set bits FFFFFFC0
    REG_0x28c   = 0x28c, Register32     # can set bits FFFFFFC0

    REG_0x290   = 0x290, Register32     # can set bits BFFF3FFF
    REG_0x294   = 0x294, Register32     # can set bits 000000FF
    REG_0x298   = 0x298, Register32     # RO? init value 07FB066F

    REG_0x2a0   = 0x2a0, Register32     # can set bits BFFF3FFF
    REG_0x2a4   = 0x2a4, Register32     # can set bits 000000FF
    REG_0x2a8   = 0x2a8, Register32     # RO? init value 037C03EE

    REG_0x2b0   = 0x2b0, Register32     # can set bits BFFF3FFF
    REG_0x2b4   = 0x2b4, Register32     # can set bits 000000FF
    REG_0x2b8   = 0x2b8, Register32     # RO? init value 04E00377

    REG_0x2c0   = 0x2c0, Register32     # can set bits BFFF3FFF
    REG_0x2c4   = 0x2c4, Register32     # can set bits 000000FF
    REG_0x2c8   = 0x2c8, Register32     # RO? init value 051C00DA

    REG_0x2d0   = 0x2d0, Register32     # can set bits FFFFFFFD, CANNOT clear 00000011
    REG_0x2d4   = 0x2d4, Register32     # can set bits 00000001
    REG_0x2d8   = 0x2d8, Register32     # can set bits FFFF0007
    REG_0x2dc   = 0x2dc, Register32     # RO? init value 07FB066F
    REG_0x2e0   = 0x2e0, Register32     # can set bits 07FF07FF

    REG_0x2f0   = 0x2f0, Register32     # can set bits FFFFFFFF

    REG_0x2f8   = 0x2f8, Register32     # can set bits FFFFFFFD, CANNOT clear 00000011
    REG_0x2fc   = 0x2fc, Register32     # can set bits 00000001
    REG_0x300   = 0x300, Register32     # can set bits FFFF0007
    REG_0x304   = 0x304, Register32     # RO? init value 037C03EE
    REG_0x308   = 0x308, Register32     # can set bits 07FF07FF

    REG_0x318   = 0x318, Register32     # can set bits FFFFFFFF

    REG_0x320   = 0x320, Register32     # can set bits FFFFFFFD, CANNOT clear 00000011
    REG_0x324   = 0x324, Register32     # can set bits 00000001
    REG_0x328   = 0x328, Register32     # can set bits FFFF0007
    REG_0x32c   = 0x32c, Register32     # RO? init value 04E00377
    REG_0x330   = 0x330, Register32     # can set bits 07FF07FF

    REG_0x340   = 0x340, Register32     # can set bits FFFFFFFF

    REG_0x350   = 0x350, Register32     # can set bits BFFF3FFF
    REG_0x354   = 0x354, Register32     # can set bits 07FF07FF
    REG_0x358   = 0x358, Register32     # can set bits FFFFFFC0
    REG_0x35c   = 0x35c, Register32     # can set bits 000003FF
    REG_0x360   = 0x360, Register32     # can set bits FFFFFFC0
    REG_0x364   = 0x364, Register32     # can set bits FFFFFFC0
    REG_0x368   = 0x368, Register32     # can set bits FFFFFFC0

    REG_0x370   = 0x370, Register32     # can set bits BFFF3FFF
    REG_0x374   = 0x374, Register32     # can set bits 07FF07FF


    QUANT_LUMA_EHQ      = irange(0x0800, 32, 4), Register32
    QUANT_LUMA_HQ       = irange(0x0880, 32, 4), Register32
    QUANT_LUMA_NQ       = irange(0x0900, 32, 4), Register32
    QUANT_LUMA_LT       = irange(0x0980, 32, 4), Register32
    QUANT_LUMA_PROXY    = irange(0x0A00, 32, 4), Register32
    QUANT_CHROMA_EHQ    = irange(0x1000, 32, 4), Register32
    QUANT_CHROMA_HQ     = irange(0x1080, 32, 4), Register32
    QUANT_CHROMA_NQ     = irange(0x1100, 32, 4), Register32
    QUANT_CHROMA_LT     = irange(0x1180, 32, 4), Register32
    QUANT_CHROMA_PROXY  = irange(0x1200, 32, 4), Register32

    # wtf, writing to this doesn't actually work? do we have to enable it?
    DC_QUANT_SCALE      = irange(0x1800, 112, 4), Register32

    REG_0x19c0  = 0x19c0, Register32    # unknown, all 1s, RO?
    REG_0x19c4  = 0x19c4, Register32    # unknown, all 1s, RO?
    REG_0x19c8  = 0x19c8, Register32    # unknown, all 1s, RO?
    REG_0x19cc  = 0x19cc, Register32    # unknown, all 1s, RO?
    REG_0x19d0  = 0x19d0, Register32    # unknown, all 1s, RO?
    REG_0x19d4  = 0x19d4, Register32    # unknown, all 1s, RO?
    REG_0x19d8  = 0x19d8, Register32    # unknown, all 1s, RO?
    REG_0x19dc  = 0x19dc, Register32    # unknown, can set bits 00000001

    # Unknown, inits to 0x12345678, can R/W
    REG_0x1A00  = 0x1a00, Register32



# SPDX-License-Identifier: MIT
from ..common import *
from ...utils import align
import construct
import struct

ISPIPCBootArgs = Struct(
    "pad" / Default(Int64ul, 0),
    "ipc_iova" / Hex(Int64ul),  # 0x1804000

    "unk0" / Hex(Int64ul),  # 0x1800000
    "unk1" / Hex(Int64ul),  # 0xe800000

    "extra_iova" / Hex(Int64ul),  # 0x1824000
    "extra_size" / Hex(Int64ul),  # 0x2200000

    "unk4" / Hex(Int32ul),  # 0x1
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),

    "pad" / Padding(0x10),

    "ipc_size" / Hex(Int32ul),  # 0x1c000
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),

    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "unk5" / Hex(Int32ul),  # 0x40
    "pad" / Default(Int32ul, 0),

    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "unk6" / Hex(Int32ul),  # 0x0 or 0x4b4c000
    "pad" / Default(Int32ul, 0),

    "pad" / Padding(0x20),

    "pad" / Default(Int32ul, 0),
    "unk7" / Hex(Int32ul),  # 0x1
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),

    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "unk_iova" / Hex(Int32ul),  # 0x18130f4

    "pad" / Padding(0xb0),

    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "unk9" / Hex(Int32ul),  # 0x3
)
assert((ISPIPCBootArgs.sizeof() == 0x180))

ISPIPCChanTableDescEntry = Struct(
    "name" / PaddedString(0x40, "utf8"),
    "type" / Int32ul,
    "src" / Int32ul,
    "num" / Int32ul,
    "pad" / Int32ul,
    "iova" / Hex(Int32ul),
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "pad" / Default(Int32ul, 0),
    "pad" / Padding(0xa0),
)
assert((ISPIPCChanTableDescEntry.sizeof() == 0x100))

class ISPChannelMessage:
    def __init__(self, arg0=0x0, arg1=0x0, arg2=0x0, arg3=0x0, arg4=0x0, arg5=0x0, arg6=0x0, arg7=0x0):
        self.count = 8
        self.arg0 = arg0; self.arg1 = arg1; self.arg2 = arg2; self.arg3 = arg3; self.arg4 = arg4; self.arg5 = arg5; self.arg6 = arg6; self.arg7 = arg7
        self.args = [self.arg0, self.arg1, self.arg2, self.arg3, self.arg4, self.arg5, self.arg6, self.arg7]
        self.data = None

    @classmethod
    def build(cls, arg0=0x0, arg1=0x0, arg2=0x0, arg3=0x0, arg4=0x0, arg5=0x0, arg6=0x0, arg7=0x0):
            return cls(arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7)

    @classmethod
    def new(cls): return cls.build()

    @classmethod
    def parse(cls, buf, index=0):
        arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7 = struct.unpack("<8q", buf)
        out = cls(arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7)
        out.data = buf
        out.index = index
        return out

    def encode(self):
            return struct.pack("<8q", *(self.arg0, self.arg1, self.arg2, self.arg3, self.arg4, self.arg5, self.arg6, self.arg7))

    def __str__(self, max_count=3):
        s = "ISP Message ["
        for n in range(max_count):
            s += "0x%x" % (getattr(self, f"arg{n}"))
            if (n < max_count-1):
                        s += ", "
        s += "]"
        return s

    def valid(self):  # rough check used for dumps
        return not (self.arg0 & 0x1)

class ISPChannel:
    """
    Ring buffers used for CPU <-> ASC communication.

    TM: TERMINAL     | src = 0, type = 2, num = 768, size = 0xc000, iova = 0x1804700
    IO: IO           | src = 1, type = 0, num =   8, size = 0x0200, iova = 0x1810700
    DG: DEBUG        | src = 1, type = 0, num =   8, size = 0x0200, iova = 0x1810900
    BR: BUF_H2T      | src = 2, type = 0, num =  64, size = 0x1000, iova = 0x1810b00
    BT: BUF_T2H      | src = 3, type = 1, num =  64, size = 0x1000, iova = 0x1811b00
    SM: SHAREDMALLOC | src = 3, type = 1, num =   8, size = 0x0200, iova = 0x1812b00
    IT: IO_T2H       | src = 3, type = 1, num =   8, size = 0x0200, iova = 0x1812d00

    There's a sticky note on my machine that says, "Host is Me, Target is Firmware".

    TM: Logs from firmware. arg0 is pointer & arg1 is the message size of log line.
    IO: To dispatch CISP_CMD_* ioctls to the firmware.
    DG: Unused.
    BR: New RX buffer pushed.
    BT: New TX buffer pushed.
    SM: Firmware requests to allocate OR free a shared mem region.
    IT: Mcache stuff when RPC is enabled. Basically unused.
    """

    def __init__(self, isp, name, _type, src, num, iova, abbrev=""):
        self.isp = isp
        self.name = name
        self.src = src
        self.type = _type
        self.num = num
        self.entry_size = 64
        self.size = self.num * self.entry_size
        self.iova = iova
        self.doorbell = 1 << self.src

        self.cursor = 0
        self.abbrev = abbrev if abbrev else self.name
        self._stfu = False

    @property
    def stfu(self): self._stfu = True

    def log(self, *args):
        if (not self._stfu):
            if (args): print("ISP: %s:" % (self.abbrev.upper()), *args)
            else: print()

    def get_iova(self, index):
            assert((index < self.num))
            return self.iova + (index * self.entry_size)

    def read_msg(self, index):
            assert((index < self.num))
            return ISPChannelMessage.parse(self.isp.ioread(self.get_iova(index), self.entry_size))

    def write_msg(self, msg, index):
            assert((index < self.num))
            self.isp.iowrite(self.get_iova(index), msg.encode())

    def read_all_msgs(self):
        ring = self.isp.ioread(self.iova, self.size)
        return [ISPChannelMessage.parse(ring[n*self.entry_size:(n+1)*self.entry_size]) for n in range(self.num)]

    def update_cursor(self):
        if (self.cursor >= (self.num - 1)):
            self.cursor = 0
        else:
            self.cursor += 1

    def dump(self):
        self.log(f'---------------- START OF {self.name} TABLE ----------------')
        for n, msg in enumerate(self.read_all_msgs()):
            if (msg.valid()):
                self.log(f'{n}: {msg}')
        self.log(f'---------------- END OF {self.name} TABLE ----------------')

    def _irq_wrap(f):
        def wrap(self, **kwargs):
            self.log()
            self.log(f'---------------- START OF {self.name} IRQ ----------------')
            x = f(self, **kwargs)
            self.log(f'---------------- END OF {self.name} IRQ ------------------')
            self.log()
            return x
        return wrap

    @_irq_wrap
    def handle_once(self, req):
        rsp = self._handle(req=req)
        if (rsp == None): return None
        self.write_msg(rsp, self.cursor)
        self.isp.regs.ISP_IRQ_DOORBELL.val = self.doorbell
        self.update_cursor()
        return rsp

    def handler(self):
        while True:
            req = self.read_msg(self.cursor)
            if ((req.arg0 & 0xf) == 0x1): break  # ack flag
            rsp = self.handle_once(req=req)
            if (rsp == None): raise RuntimeError("IRQ stuck")

    @_irq_wrap
    def send_once(self, req):
        self.log(f'TX: REQ: {req}')
        self.write_msg(req, self.cursor)

        rsp = None
        self.isp.regs.ISP_IRQ_DOORBELL.val = self.doorbell
        while True:
            rsp = self.read_msg(self.cursor)
            if (rsp.arg0 == (req.arg0 | 0x1)): break  # ack flag
            self.isp.table.sharedmalloc.handler()
        if (rsp == None): return None

        self.log(f'RX: RSP: {rsp}')
        self.isp.regs.ISP_IRQ_ACK.val = self.isp.regs.ISP_IRQ_INTERRUPT.val
        self.update_cursor()
        return rsp

    def send(self, req):
        rsp = self.send_once(req=req)
        if (rsp == None): raise RuntimeError("ASC never acked. IRQ stuck.")
        return rsp



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1.fw.smc import SMCClient

smc_addr = u.adt["arm-io/smc"].get_reg(0)[0]
smc = SMCClient(u, smc_addr, None)
smc.verbose = 3

smc.start()
smc.start_ep(0x20)

run_shell(globals(), msg="Have fun!")

smc.stop()



# SPDX-License-Identifier: MIT
from ..utils import *

__all__ = ["SPIRegs"]

class R_CTRL(Register32):
    RX_FIFO_RESET   = 3
    TX_FIFO_RESET   = 2
    RUN             = 0

class R_CFG(Register32):
    # impl: 002fb1e6
    IE_TX_COMPLETE  = 21
    b19             = 19
    FIFO_THRESH     = 18, 17
        # 0 = 8 bytes
        # 1 = 4 bytes
        # 2 = 1 byte
        # 3 = disabled
    WORD_SIZE       = 16, 15
        # 0 = 8bit
        # 1 = 16bit
        # 2 = 32bit
    LSB_FIRST       = 13
    b12             = 12
    IE_RX_THRESH    = 8
    IE_RX_COMPLETE  = 7
    MODE            = 6, 5
        # 0 = polled
        # 1 = irq
    CPOL            = 2
    CPHA            = 1

class R_STATUS(Register32):
    TX_COMPLETE     = 22
    TXRX_THRESH     = 1     # updated if MODE == 1
    RX_COMPLETE     = 0

class R_PIN(Register32):
    CS              = 1
    KEEP_MOSI       = 0

class R_CLKDIV(Register32):
    DIVIDER         = 10, 0 # SPI freq = CLK / (DIVIDER + 1)

class R_INTER_DELAY(Register32):
    DELAY           = 15, 0

class R_FIFOSTAT(Register32):
    LEVEL_RX        = 31, 24
    RX_EMPTY        = 20
    LEVEL_TX        = 15, 8
    TX_FULL         = 4

class R_IRQ_XFER(Register32):
    TX_XFER_DONE    = 1
    RX_XFER_DONE    = 0

class R_IRQ_FIFO(Register32):
    TX_OVERFLOW     = 17
    RX_UNDERRUN     = 16
    TX_EMPTY        = 9
    RX_FULL         = 8
    TX_THRESH       = 5
    RX_THRESH       = 4

class R_XFSTATUS(Register32):
    SR_FULL         = 26
    SHIFTING        = 20
    STATE           = 17, 16
    UNK             = 0

class R_DIVSTATUS(Register32):
    COUNT2          = 31, 16
    COUNT1          = 15, 0

class R_SHIFTCFG(Register32):
    OVERRIDE_CS     = 24
    BITS            = 21, 16
    RX_ENABLE       = 11
    TX_ENABLE       = 10
    CS_AS_DATA      = 9
    AND_CLK_DATA    = 8
    #?              = 2  # needs to be 1 for RX to not break
    CS_ENABLE       = 1
    CLK_ENABLE      = 0

class R_PINCFG(Register32):
    MOSI_INIT_VAL   = 10
    CS_INIT_VAL     = 9
    CLK_INIT_VAL    = 8
    KEEP_MOSI       = 2
    KEEP_CS         = 1
    KEEP_CLK        = 0

class R_DELAY(Register32):
    DELAY           = 31, 16
    MOSI_VAL        = 12
    CS_VAL          = 10
    SCK_VAL         = 8
    SET_MOSI        = 6
    SET_CS          = 5
    SET_SCK         = 4
    NO_INTERBYTE    = 1
    ENABLE          = 0

class R_SCKCFG(Register32):
    PERIOD          = 31, 16
    PHASE1          = 9
    PHASE0          = 8
    RESET_TO_IDLE   = 4

class R_SCKPHASES(Register32):
    PHASE1_START   = 31, 16
    PHASE0_START   = 15, 0

class SPIRegs(RegMap):
    CTRL        = 0x00, R_CTRL
    CFG         = 0x04, R_CFG
    STATUS      = 0x08, R_STATUS
    PIN         = 0x0C, R_PIN
    TXDATA      = 0x10, Register32
    RXDATA      = 0x20, Register32
    CLKDIV      = 0x30, R_CLKDIV
    RXCNT       = 0x34, Register32
    INTER_DELAY = 0x38, R_INTER_DELAY
    TXCNT       = 0x4C, Register32
    FIFOSTAT    = 0x10C, R_FIFOSTAT

    IE_XFER     = 0x130, R_IRQ_XFER
    IF_XFER     = 0x134, R_IRQ_XFER
    IE_FIFO     = 0x138, R_IRQ_FIFO
    IF_FIFO     = 0x13c, R_IRQ_FIFO

    SHIFTCFG    = 0x150, R_SHIFTCFG
    PINCFG      = 0x154, R_PINCFG

    DELAY_PRE   = 0x160, R_DELAY
    SCKCFG      = 0x164, R_SCKCFG
    DELAY_POST  = 0x168, R_DELAY

    SCKPHASES  = 0x180, R_SCKPHASES

    UNK_PHASE   = 0x18c, Register32 # probably MISO sample point

    XFSTATUS    = 0x1c0, R_XFSTATUS
    DIVSTATUS   = 0x1e0, R_DIVSTATUS



# SPDX-License-Identifier: MIT
from .base import *
from ...utils import *

class KDebugMessage(Register64):
    TYPE    = 55, 48

class KDebugGetBufMessage(KDebugMessage):
    TYPE    = 55, 48, Constant(1)
    COUNT   = 47, 0

class KDebugPreallocBuf1Message(KDebugMessage):
    TYPE    = 55, 48, Constant(2)
    DVA     = 47, 12
    FLAGS   = 11, 0

class KDebugPreallocBuf2Message(KDebugMessage):
    TYPE    = 55, 48, Constant(3)
    DVA     = 47, 0

class KDebugSendBufMessage(KDebugMessage):
    TYPE    = 55, 48
    DVA     = 47, 0

class KDebugStart(KDebugMessage):
    TYPE    = 55, 48, Constant(8)

class ASCKDebugEndpoint(ASCBaseEndpoint):
    SHORT = "kdebug"
    BASE_MESSAGE = KDebugMessage

    @msg_handler(1, KDebugGetBufMessage)
    def GetBuf(self, msg):
        size = align_up(msg.COUNT * 0x20, 0x4000)
        self.iobuffer0, self.iobuffer0_iova = self.asc.ioalloc(size)
        self.send(KDebugSendBufMessage(TYPE=1, DVA=self.iobuffer0_iova))

        self.iobuffer1, self.iobuffer1_iova = self.asc.ioalloc(0x2000)
        self.send(KDebugSendBufMessage(TYPE=2, DVA=self.iobuffer1_iova))
        return True

    @msg_handler(2, KDebugPreallocBuf1Message)
    def SetBuf1(self, msg):
        #self.send(KDebugSendBufMessage(TYPE=1, DVA=msg.DVA))
        return True

    @msg_handler(3, KDebugPreallocBuf2Message)
    def SetBuf2(self, msg):
        #self.send(KDebugSendBufMessage(TYPE=2, DVA=msg.DVA))
        return True

    def start(self):
        self.iobuffer0 = None
        self.iobuffer1 = None
        self.iobuffer0_iova = None
        self.iobuffer1_iova = None
        self.send(KDebugStart())



# SPDX-License-Identifier: MIT

from enum import Enum

from m1n1.trace.i2c import I2CTracer, I2CDevTracer

class HpmTracer(I2CDevTracer):
    class State(Enum):
        UNKNOWN = 0
        REQUEST = 1
        WRITE = 2
        READ = 3

    def __init__(self, addr=128, name=None, verbose=True):
        print(f"CD3217Tracer.__init__(addr={addr}, name={name}, verbose={verbose})")
        super().__init__(addr, name, verbose)
        self.reset()

    def reset(self):
        self.reg = None
        self.state = CD3217Tracer.State.UNKNOWN
        self.length = None
        self.data = []

    def start(self, addr, read):
        if addr != self.addr:
            return

        if self.state == CD3217Tracer.State.UNKNOWN:
            if read:
                self.state = CD3217Tracer.State.READ
            else:
                self.state = CD3217Tracer.State.REQUEST
        elif self.state == CD3217Tracer.State.REQUEST and read:
            pass
        else:
            self.log(f"unexpected state in start(read={read}): state:{self.state} reg:{self.reg} data:{self.data}")

    def stop(self):
        if self.state == CD3217Tracer.State.REQUEST and len(self.data) == 0:
            return

        msg = f"Txn: {self.addr:02x}."
        if self.state == CD3217Tracer.State.REQUEST:
            msg += f"r [{self.reg:02x}]"
        elif self.state == CD3217Tracer.State.WRITE:
            msg += f"w [{self.reg:02x}]"
        elif self.state == CD3217Tracer.State.READ:
            msg += f"r [xx]"
        else:
            self.log(f"unexpected state in stop(): state:{self.state} reg:{self.reg} data:{self.data}")
            self.reset()
            return

        # only for debugging as some mismatches are expected as
        # cd3217 seems to report the register size and not the number
        # of requested bytes (or I2CDevTracer truncates reads).
        #if self.length is not None and self.length > len(self.data):
        #    self.log(f"length {self.length:02x} mismatch received data: {len(self.data):02x}")

        for data in self.data:
            msg += f" {data:02x}"

        self.log(msg)
        self.reset()

    def read(self, data):
        self.data.append(data)

    def write(self, data):
        if self.reg is None:
            self.reg = data
        elif self.length is None:
            self.length = data
            self.state = CD3217Tracer.State.WRITE
        else:
            self.data.append(data)

class CD3217Tracer(HpmTracer):
    def read(self, data):
        if self.length is None:
            self.length = data - 1
        else:
            self.data.append(data)


i2c_tracers = {}

for node in hv.adt["/arm-io"]:
    if not node.name.startswith("i2c"):
        continue

    n = int(node.name[3:])
    bus = I2CTracer(hv, f"/arm-io/{node.name}")

    for mngr_node in node:
        if "compatible" not in mngr_node._properties: # thanks Apple
            continue

        if mngr_node.compatible[0] != "usbc,manager":
            continue

        addr = mngr_node.reg[0] & 0xff
        bus.add_device(addr, HpmTracer(addr=addr, name=mngr_node.name))

        for devnode in mngr_node:

            dcls = {
                "usbc,cd3217": CD3217Tracer,
            }.get(devnode.compatible[0], None)
            if dcls:
                addr = devnode.hpm_iic_addr & 0xff
                bus.add_device(addr, dcls(addr=addr, name=devnode.name))

    if len(bus.state.devices) > 1:
        i2c_tracers[n] = bus

for bus in i2c_tracers.values():
    bus.start()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1 import adt
from m1n1.setup import *
from m1n1.hw.nco import NCO

dt = u.adt

pmgr = dt["/arm-io/pmgr"]

dev_by_id = {dev.id: dev for dev in pmgr.devices}
pd_by_id = {pd.id: pd for pd in pmgr.power_domains}
clk_by_id = {clk.id: clk for clk in pmgr.clocks}

print("=== PS Regs ===")
for i, r in enumerate(pmgr.ps_regs):
    print(f" #{i:2d} reg: {r.reg} off: {r.offset:05x} mask:{r.mask:08x}")

print()
print("=== Perf Regs ===")
for i, r in enumerate(pmgr.perf_regs):
    print(f" #{i:2d} reg: {r.reg} off: {r.offset:05x} size:{r.size:05x} unk:{r.unk:08x}")

#print()
#print("=== PWR Gate Regs ===")
#for i, r in enumerate(pmgr.pwrgate_regs):
    #print(f" #{i:2d} reg: {r.reg} off: {r.offset:05x} mask:{r.mask:08x} unk:{r.unk:08x}")

clock_users = {}
dev_users = {}

for dev in dt["/arm-io"]:
    if hasattr(dev, "clock_ids") and dev.clock_ids:
        for i, clk in enumerate(dev.clock_ids):
            clock_users.setdefault(clk, []).append(f"{dev._path}.clk[{i}]")
    if hasattr(dev, "clock_gates") and dev.clock_gates:
        for i, pdev in enumerate(dev.clock_gates):
            dev_users.setdefault(pdev, []).append(f"{dev._path}.clkgate[{i}]")
    if hasattr(dev, "power_gates") and dev.power_gates:
        for i, pdev in enumerate(dev.power_gates):
            dev_users.setdefault(pdev, []).append(f"{dev._path}.pwrgate[{i}]")

print()
print("=== Devices ===")
for i, dev in enumerate(pmgr.devices):
    flags = ", ".join(k for k in dev.flags if k[0] != "_" and dev.flags[k])
    s = f" #{i:3d} {dev.name:20s} id: {dev.id:3d} psreg: {dev.psreg:2d}:{dev.psidx:2d} "
    s += f" flags: {flags:24s} unk1_0: {dev.unk1_0} unk1_1: {dev.unk1_1} unk1_2: {dev.unk1_2} "
    s += f" perf_reg: {dev.perf_block}:{dev.perf_idx:#04x} unk3: {dev.unk3:3d} {dev.unk2_0:2d} {dev.ps_cfg16:2d} {dev.unk2_3:3d}"

    if not dev.flags.no_ps:
        ps = pmgr.ps_regs[dev.psreg]
        addr = pmgr.get_reg(ps.reg)[0] + ps.offset + dev.psidx * 8
        val = p.read32(addr)
        s += f" @ {addr:#x} = {val:#010x}"
    else:
        s += f" @                         "
    if dev.pd:
        pd = pd_by_id[dev.pd]
        s += f" pd: {pd.name:20s}"
    else:
        s += "                         "
    if any(dev.parents):
        s += " parents: " + ", ".join(dev_by_id[idx].name if idx in dev_by_id else f"#{idx}" for idx in dev.parents if idx)
    print(s)
    for i in dev_users.get(dev.id, []):
        print(f"  User: {i}")

print()
print("=== Clocks ===")
for i, clk in enumerate(pmgr.clocks):
    perf = pmgr.perf_regs[clk.perf_block]
    reg = pmgr.get_reg(perf.reg)[0] + 0x100 + clk.perf_idx * 0x10
    print(f" #{i:3d} {clk.name:20s} id: {clk.id:3d} reg:{clk.perf_block}:{clk.perf_idx:#4x} ({reg:#x}) {clk.unk:#x}")

print()
print("=== Power Domains ===")
for i, pd in enumerate(pmgr.power_domains):
    perf = pmgr.perf_regs[pd.perf_block]
    reg = pmgr.get_reg(perf.reg)[0] + 0x100 + pd.perf_idx * 0x10
    print(f" #{i:3d} {pd.name:20s} id: {pd.id:3d} reg:{pd.perf_block}:{pd.perf_idx:#4x} ({reg:#x})")

print()
print("=== Events ===")
for i, ev in enumerate(pmgr.events):
    perf = pmgr.perf_regs[ev.perf_block]
    reg = pmgr.get_reg(perf.reg)[0] + 0x100 + ev.perf_idx * 0x10
    v = f" #{i:3d} {ev.name:20s} unk:{ev.unk1:#3x}/{ev.unk2}/{ev.unk3} id: {ev.id:3d} reg:{ev.perf_block}:{ev.perf_idx:#4x} ({reg:#x})"
    if ev.perf2_idx:
        perf2 = pmgr.perf_regs[ev.perf2_block]
        reg2 = pmgr.get_reg(perf2.reg)[0] + 0x100 + ev.perf2_idx * 0x10
        v += f" reg2:{ev.perf2_block}:{ev.perf2_idx:#4x} ({reg2:#x})"
    print(v)

arm_io = dt["/arm-io"]

print()
print("=== Fixed clocks ===")
for clk in range(256):
    users = clock_users.get(clk, [])
    if users:
        print(f" #{clk}")
        for j in users:
            print(f"  User: {j}")

print()
print("=== Boot clocks ===")
for i, (freq, reg, nclk) in enumerate(zip(arm_io.clock_frequencies,
                                          arm_io.clock_frequencies_regs,
                                          arm_io.clock_frequencies_nclk)):
    v = ""
    clk_type = reg >> 56
    reg = reg & 0xFFFFFFFFFFFFFF
    
    if clk_type == 0x9c:
        v = f"fixed: {reg}"
    elif clk_type in (0xa0, 0xa1, 0xa4, 0xa5):
        v = f"regval: {p.read32(reg):#x}"
    elif clk_type == 0xa8:
        regvals = [p.read32(reg+off*4) for off in range(5)]
        try:
            # we are using the freq value from ADT as fin of NCO,
            # that's not exactly correct
            nco_freq = NCO.calc_rate(freq, regvals)
        except ValueError as e:
            nco_freq = 0

        v = f"nco: (calculated rate: {nco_freq:d}) "
        for val in regvals:
            v += f"{val:#x} "
        
    
    print(f"#{i:3}: {freq:10d} {nclk} {clk_type:#x}/{reg:#x}: {v}")
    for j in clock_users.get(i + 256, []):
        print(f"  User: {j}")



# SPDX-License-Identifier: MIT
import time
from construct import *
from ..afk.epic import *
from .ipc import *

# spuapp
class AOPSPUAppService(EPICService):
    NAME = "SPUApp"
    SHORT = "spuapp"

class AOPSPUAppI2CService(EPICService):
    NAME = "i2c"
    SHORT = "i2c"

class AOPSPUAppEndpoint(EPICEndpoint):
    SHORT = "spuapp"
    SERVICES = [
        AOPSPUAppService,
        AOPSPUAppI2CService,
    ]

# accel
class AOPAccelService(EPICService):
    NAME = "accel"
    SHORT = "accel"

class AOPAccelEndpoint(EPICEndpoint):
    SHORT = "accel"
    SERVICES = [
        AOPAccelService,
    ]

# gyro
class AOPGyroService(EPICService):
    NAME = "gyro"
    SHORT = "gyro"

class AOPGyroEndpoint(EPICEndpoint):
    SHORT = "gyro"
    SERVICES = [
        AOPGyroService,
    ]

    def start_queues(self):
        pass  # don't init gyro ep (we don't have one)

# als
class AOPALSService(EPICService):
    NAME = "als"
    SHORT = "als"

    @report_handler(0xc4, ALSLuxReport)
    def handle_lux(self, seq, fd, rep):
        self.log(rep)
        return True

class AOPALSEndpoint(EPICEndpoint):
    SHORT = "als"
    SERVICES = [
        AOPALSService,
    ]

    def send_notify(self, call, chan="als"):
        return super(AOPALSEndpoint, self).send_notify(chan, call)

    def send_roundtrip(self, call, chan="als"):
        return super(AOPALSEndpoint, self).send_roundtrip(chan, call)

    def send_cmd(self, call, chan="als"):
        return super(AOPALSEndpoint, self).send_cmd(chan, call)

# wakehint
class AOPWakehintService(EPICService):
    NAME = "wakehint"
    SHORT = "wakehint"

class AOPWakehintEndpoint(EPICEndpoint):
    SHORT = "wakehint"
    SERVICES = [
        AOPWakehintService,
    ]

# unk26
class AOPUNK26Service(EPICService):
    NAME = "unk26"
    SHORT = "unk26"

class AOPUNK26Endpoint(EPICEndpoint):
    SHORT = "unk26"
    SERVICES = [
        AOPUNK26Service,
    ]

# audio
class AOPAudioService(EPICService):
    NAME = "aop-audio"
    SHORT = "audio"

class AOPAudioEndpoint(EPICEndpoint):
    SHORT = "audio"
    SERVICES = [
        AOPAudioService,
    ]

    def send_ipc(self, data, chan="aop-audio", **kwargs):
        return super(AOPAudioEndpoint, self).send_ipc(data, **kwargs)

    def send_notify(self, call, chan="aop-audio", **kwargs):
        return super(AOPAudioEndpoint, self).send_notify(chan, call, **kwargs)

    def send_roundtrip(self, call, chan="aop-audio", **kwargs):
        return super(AOPAudioEndpoint, self).send_roundtrip(chan, call, **kwargs)

    def send_cmd(self, call, chan="aop-audio", **kwargs):
        return super(AOPAudioEndpoint, self).send_cmd(chan, call, **kwargs)

    def send_notifycmd(self, type, data, chan="aop-audio", **kwargs):
        return super(AOPAudioEndpoint, self).send_notifycmd(chan, type, data, **kwargs)

class AOPVoiceTriggerService(EPICService):
    NAME = "aop-voicetrigger"
    SHORT = "voicetrigger"

class AOPVoiceTriggerEndpoint(EPICEndpoint):
    SHORT = "voicetrigger"
    SERVICES = [
        AOPVoiceTriggerService,
    ]

    def send_notify(self, call, chan="aop-voicetrigger", **kwargs):
        return super(AOPVoiceTriggerEndpoint, self).send_notify(chan, call, **kwargs)

    def send_roundtrip(self, call, chan="aop-voicetrigger", **kwargs):
        return super(AOPVoiceTriggerEndpoint, self).send_roundtrip(chan, call, **kwargs)

    def send_cmd(self, type, data, chan="aop-voicetrigger", **kwargs):
        return super(AOPVoiceTriggerEndpoint, self).send_cmd(chan, type, data, **kwargs)

    def send_notifycmd(self, type, data, chan="aop-voicetrigger", **kwargs):
        return super(AOPVoiceTriggerEndpoint, self).send_notifycmd(chan, type, data, **kwargs)
